import re

from io import BytesIO
from typing import ByteString, Iterator
from warnings import warn

from tivars.models import *
from tivars.tokenizer import TokenizedString
from ..data import *
from ..var import TIEntry
from .complex import ComplexEntry
from .real import RealEntry


class ListName(TokenizedString):
    """
    Converter for the name section of lists

    List names can be `L1` - `L6` or a string of five alphanumeric characters that do not start with a digit.
    The special name and token `IDList` is also used (but is planned to be relegated to a separate type).
    """

    _T = str

    @classmethod
    def get(cls, data: bytes, **kwargs) -> _T:
        """
        Converts `bytes` -> `str` as done by the memory viewer

        :param data: The raw bytes to convert
        :return: The list name contained in `data`
        """

        if data[0] == 0x5D:
            if data[1] < 6:
                return super().get(data)

            elif data[1] == 0x40:
                return "IDList"

            data = data[1:]

        return super().get(data)

    @classmethod
    def set(cls, value: _T, **kwargs) -> bytes:
        """
        Converts `str` -> `bytes` to match appearance in the memory viewer

        :param value: The value to convert
        :return: The name encoding of `value`
        """

        varname = value[:7].upper()
        varname = re.sub(r"(\u03b8|\u0398|\u03F4|\u1DBF)", "θ", varname)
        varname = re.sub(r"]", "|L", varname)
        varname = re.sub(r"[^θa-zA-Z0-9]", "", varname)

        if "IDList" in varname:
            return b']@'

        elif varname.startswith("|L"):
            return super().set(varname[-5:])

        else:
            return super().set(varname[:2])


class ListEntry(TIEntry):
    """
    Base class for all list entries

    A list entry is a one-dimensional array of `RealEntry` or `ComplexEntry` elements.
    Exact types are supported.
    """

    _E = TIEntry

    versions = [0x10, 0x0B, 0x00]

    min_data_length = 2

    def __init__(self, init=None, *,
                 for_flash: bool = True, name: str = "L1",
                 version: bytes = None, archived: bool = None,
                 data: ByteString = None):

        super().__init__(init, for_flash=for_flash, name=name, version=version, archived=archived, data=data)

    def __format__(self, format_spec: str) -> str:
        match format_spec:
            case "":
                return "[" + ", ".join(format(entry, format_spec) for entry in self.list()) + "]"
            case "t":
                return "{" + ",".join(format(entry, 't') for entry in self.list()) + "}"
            case _:
                return super().__format__(format_spec)

    def __iter__(self) -> Iterator[_E]:
        return iter(self.list())

    @Section(8, ListName)
    def name(self) -> str:
        """
        The name of the entry

        Names must be 1 to 5 characters in length.
        The name can include any characters A-Z, 0-9, or θ.
        The name cannot start with a digit; for these lists, use `L1` - `L6` instead.
        """

    @Section()
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, Integer)[0:2]
    def length(self, value) -> int:
        """
        The length of the list

        TI-OS imposes a limit of 999.
        """

        if value > 999:
            warn(f"The list is too long ({value} > 999).",
                 UserWarning)

        return value

    @View(calc_data, Bytes)[2:]
    def data(self) -> bytearray:
        pass

    def get_min_os(self, data: bytes = None) -> OsVersion:
        it = zip(*[iter(data or self.data)] * RealEntry.min_data_length)
        return max(map(self._E().get_min_os, it), default=OsVersions.INITIAL)

    def get_version(self, data: bytes = None) -> int:
        it = zip(*[iter(data or self.data)] * RealEntry.min_data_length)
        version = max(map(self._E().get_version, it), default=0x00)

        if version > 0x1B:
            return 0x10

        elif version == 0x1B:
            return 0x0B

        else:
            return 0x00

    def supported_by(self, model: TIModel) -> bool:
        return super().supported_by(model) and (self.get_version() <= 0x0B or model.has(TIFeature.ExactMath))

    @Loader[ByteString, BytesIO]
    def load_bytes(self, data: bytes | BytesIO):
        super().load_bytes(data)

        if self.data_length // self._E.min_data_length != self.length:
            warn(f"The list has an unexpected length "
                 f"(expected {self.data_length // self._E.min_data_length}, got {self.length}).",
                 BytesWarning)

    @Loader[list]
    def load_list(self, lst: list[_E]):
        """
        Loads a `list` into this list

        :param lst: The list to load
        """

        self.load_bytes(int.to_bytes(len(lst), 2, 'little') + b''.join(entry.calc_data for entry in lst))

    def list(self) -> list[_E]:
        """
        :return: A `list` of the elements in this list
        """

        it = zip(*[iter(self.data)] * self._E.min_data_length)
        return [self._E(for_flash=self.meta_length > TIEntry.base_meta_length, data=bytes(data)) for data in it]

    @Loader[str]
    def load_string(self, string: str):
        lst = []

        for string in ''.join(string.strip("[]{}").split()).split(","):
            entry = self._E()
            entry.load_string(string)
            lst.append(entry)

        self.load_list(lst)

    def string(self) -> str:
        return format(self, "")


class TIRealList(ListEntry, register=True):
    _E = RealEntry

    extensions = {
        None: "8xl",
        TI_82: "82l",
        TI_83: "83l",
        TI_83P: "8xl",
    }

    _type_id = 0x01


class TIComplexList(ListEntry, register=True):
    _E = ComplexEntry

    extensions = {
        None: "8xl",
        TI_83: "83l",
        TI_83P: "8xl",
    }

    _type_id = 0x0D


__all__ = ["TIRealList", "TIComplexList"]



from tivars.models import *
from ..var import SizedEntry


class TIAppVar(SizedEntry, register=True):
    """
    Parser for application variables (appvars)

    An appvar is a sized block of data, to be interpreted by the corresponding app as desired.
    Certain common formats, such as compiled Python modules, are denoted by a header at the start of the data block.
    """

    flash_only = True

    extensions = {
        None: "8xv",
        TI_83P: "8xv",
    }

    _type_id = 0x15

    def get_min_os(self, data: bytes = None) -> OsVersion:
        return TI_83P.OS()


__all__ = ["TIAppVar"]



from typing import ByteString, Iterator
from warnings import warn

from tivars.models import *
from tivars.tokenizer import TokenizedString
from ..data import *
from ..var import SizedEntry

RGB = tuple[int, int, int]


class L1(Converter):
    """
    Converter for black-and-white pixels packed eight-per-byte
    """

    _T = tuple[int, ...]

    @classmethod
    def get(cls, data: bytes, **kwargs) -> _T:
        """
        Converts `bytes` -> `tuple[int, ...]`

        :param data: The raw bytes to convert
        :return: A tuple of eight integers in {0, 255} corresponding to the bits of `data`
        """

        return tuple(255 * (1 - int(bit)) for bit in f"{data[0]:08b}")

    @classmethod
    def set(cls, value: _T, **kwargs) -> bytes:
        """
        Converts  `tuple[int, ...]` -> `bytes`

        :param value: The value to convert
        :return: The values in `value` joined into a single byte
        """

        return bytes([int("".join("0" if bit else "1" for bit in value), 2)])


class RGBPalette(Converter):
    """
    Converter for color pixels indexed by the TI palette
    """

    _T = tuple[RGB, RGB]

    Blue = (0, 0, 255)
    Red = (255, 0, 0)
    Black = (0, 0, 0)
    Magenta = (255, 0, 255)
    Green = (0, 158, 0)
    Orange = (255, 142, 32)
    Brown = (182, 32, 0)
    Navy = (0, 0, 134)
    LtBlue = (0, 146, 255)
    Yellow = (255, 255, 0)
    White = (255, 255, 255)
    LtGray = (231, 227, 231)
    MedGray = (199, 195, 199)
    Gray = (142, 138, 142)
    DarkGray = (81, 85, 81)

    palette = [White, Blue, Red, Black, Magenta, Green, Orange, Brown, Navy, LtBlue, Yellow,
               White, LtGray, MedGray, Gray, DarkGray]

    @classmethod
    def nearest(cls, r: int, g: int, b: int) -> RGB:
        """
        Finds the nearest palette color to the given RGB components under the Euclidean metric

        :param r: The red component
        :param g: The green component
        :param b: The blue component
        :return: The RGB components of the nearest palette color
        """

        nearest = min(cls.palette, key=lambda x: (x[0] - r) ** 2 + (x[1] - g) ** 2 + (x[2] - b) ** 2)
        if nearest != (r, g, b):
            warn(f"The pixel {(r, g, b)} is not contained in the palette; using {nearest} as an approximation.",
                 UserWarning)

        return nearest

    @classmethod
    def get(cls, data: bytes, **kwargs) -> _T:
        """
        Converts `bytes` -> `tuple[RGB, RGB]`

        :param data: The raw bytes to convert
        :return: The pair of RGB values from the palette indexed by the nibbles in `data`
        """

        return cls.palette[data[0] >> 4], cls.palette[data[0] & 15]

    @classmethod
    def set(cls, value: _T, **kwargs) -> bytes:
        """
        Converts  `tuple[RGB, RBG]` -> `bytes`

        :param value: The value to convert
        :return: The byte formed by finding the indices of the colors nearest to the RGB values in `value`
        """

        return bytes([(cls.palette.index(cls.nearest(*value[0])) << 4) + cls.palette.index(cls.nearest(*value[1]))])


class RGB565(Converter):
    """
    Converter for color pixels stored in RGB565 format
    """

    _T = RGB

    @classmethod
    def get(cls, data: bytes, **kwargs) -> _T:
        """
        Converts `bytes` -> `RGB`

        :param data: The raw bytes to convert
        :return: The RGB value stored in the two bytes of `data`
        """

        return (
            (data[1] >> 3) * 255 // 31,
            (((data[1] & 7) << 3) | (data[0] >> 5)) * 255 // 63,
            (data[0] & 31) * 255 // 31
        )

    @classmethod
    def set(cls, value: _T, **kwargs) -> bytes:
        """
        Converts  `RGB` -> `bytes`

        :param value: The value to convert
        :return: The bytes formed by concatenating the RGB components of `value` in 565 format
        """

        return bytes([((value[1] >> 2 & 7) << 5) | value[2] >> 3]) + bytes([(value[0] & ~7) | (value[1] >> 5)])


class PictureEntry(SizedEntry):
    """
    Base class for all picture types

    A picture or image is stored as a stream of pixels in some encoding format.
    """

    width = 0
    """
    The width of the picture
    """

    height = 0
    """
    The height of the picture
    """

    data_width = width
    """
    The width of the picture while stored as data
    """

    data_height = height
    """
    The height of the picture while stored as data
    """

    pil_mode = None
    """
    The mode used by PIL images for this image type
    """

    pixel_type = None
    """
    The type of a single pixel
    """

    has_color = True
    """
    Whether this picture has color
    """

    def __iter__(self) -> Iterator[pixel_type]:
        for row in self.array():
            for col in row:
                yield col

    @Loader[list, ]
    def load_array(self, arr: list[list[pixel_type]]):
        """
        Loads a two-dimensional `list` of pixels into this picture

        :param arr: The array to load
        """

        raise NotImplementedError

    def array(self) -> list[list[pixel_type]]:
        """
        :return: A two-dimensional `list` of the pixels in this picture
        """

        raise NotImplementedError

    def coerce(self):
        match self.length + 2:
            case TIMonoPicture.min_data_length: self.__class__ = TIMonoPicture
            case TIPicture.min_data_length: self.__class__ = TIPicture
            case TIImage.min_data_length: self.__class__ = TIImage
            case _: warn(f"Picture has unexpected length ({self.length}).",
                         BytesWarning)


class TIMonoPicture(PictureEntry):
    """
    Parser for monochromatic pictures

    A `TIMonoPicture` is a 96 x 63 grid of black or white pixels, stored as 8 pixels per byte.
    """

    extensions = {
        None: "8xi",
        TI_83P: "8xi",
        TI_84PCSE: ""
    }

    min_data_length = 758

    width = 96
    height = 63

    data_width = width // 8
    data_height = height
    data_offset = 0

    pil_mode = "L"
    pixel_type = int
    has_color = False

    _type_id = 0x07

    def __init__(self, init=None, *,
                 for_flash: bool = True, name: str = "Pic1",
                 version: bytes = None, archived: bool = None,
                 data: ByteString = None):

        super().__init__(init, for_flash=for_flash, name=name, version=version, archived=archived, data=data)

    def get_min_os(self, data: bytes = None) -> OsVersion:
        return TI_83P.OS()

    @Loader[list]
    def load_array(self, arr: list[list[pixel_type]]):
        self.data = b''.join(L1.set(entry) for row in arr for entry in zip(*[iter(row)] * 8, strict=True))

    def array(self) -> list[list[pixel_type]]:
        return [[bw for col in range(self.data_width)
                 for bw in L1.get(self.data[self.data_width * row + col:][:1])]
                for row in range(self.data_height)]


class TIPicture(PictureEntry, register=True):
    """
    Parser for color pictures

    A `TIPicture` is a 266 x 165 grid of pixels which can take on any of the 15 standard colors.
    Each pixel comprises a single nibble, which stores an (offset) index into the color palette.
    """

    flash_only = True

    versions = [0x0A]

    extensions = {
        None: "8ci",
        TI_84PCSE: "8ci"
    }

    min_data_length = 21947

    width = 266
    height = 165

    data_width = width // 2
    data_height = height

    pil_mode = "RGB"
    pixel_type = RGB

    _type_id = 0x07

    def __init__(self, init=None, *,
                 for_flash: bool = True, name: str = "Pic1",
                 version: bytes = None, archived: bool = None,
                 data: ByteString = None):

        super().__init__(init, for_flash=for_flash, name=name, version=version, archived=archived, data=data)

    def get_min_os(self, data: bytes = None) -> OsVersion:
        return TI_84PCSE.OS()

    @Loader[list]
    def load_array(self, arr: list[list[pixel_type]]):
        self.data = b''.join(RGBPalette.set(entry) for row in arr for entry in zip(row[::2], row[1::2]))

    def array(self) -> list[list[pixel_type]]:
        return [[rgb for col in range(self.data_width)
                 for rgb in RGBPalette.get(self.data[self.data_width * row + col:][:1])]
                for row in range(self.data_height)]


# Workaround until the token sheets are updated
class ImageName(TokenizedString):
    _T = str

    @classmethod
    def get(cls, data: bytes, **kwargs) -> _T:
        return f"Image{data[1] + 1}"

    @classmethod
    def set(cls, value: _T, **kwargs) -> bytes:
        return b"\x3C" + bytes([int(value[-1], 16) - 1])


class TIImage(PictureEntry, register=True):
    """
    Parser for color images

    A `TIImage` is a 133 x 83 grid of pixels which can take on any value in the full color space.
    Each pixel comprises two bytes, stored in RGB565 format.

    On-calc, each pixel is duplicated in both directions to fill the screen.
    """

    flash_only = True

    versions = [0x00]

    extensions = {
        None: "8ca",
        TI_84PCSE: "8ca",
    }

    min_data_length = 22247

    width = 133
    height = 83

    data_width = 2 * width + 2
    data_height = height

    pil_mode = "RGB"
    pixel_type = RGB

    leading_bytes = b'\x81'

    _type_id = 0x1A

    def __init__(self, init=None, *,
                 for_flash: bool = True, name: str = "Image1",
                 version: bytes = None, archived: bool = None,
                 data: ByteString = None):

        super().__init__(init, for_flash=for_flash, name=name, version=version, archived=archived, data=data)

    @Section(8, ImageName)
    def name(self) -> str:
        """
        The name of the entry

        Must be one of the image names: Image1 - Image0
        """

    @Section()
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, Bytes)[2:3]
    def image_magic(self) -> bytes:
        """
        Magic identifying the file as an image

        This value is always `0x81`.
        """

    @View(calc_data, SizedData)[3:]
    def data(self) -> bytearray:
        pass

    def get_min_os(self, data: bytes = None) -> OsVersion:
        return TI_84PCSE.OS()

    @Loader[list]
    def load_array(self, arr: list[list[pixel_type]]):
        self.data = b''.join(RGB565.set(entry) for row in arr[::-1] for entry in row + [(0, 0, 0)])

    def array(self) -> list[list[pixel_type]]:
        return [[RGB565.get(self.data[self.data_width * row + col:][:2])
                 for col in range(0, self.data_width - 2, 2)]
                for row in range(self.data_height)][::-1]


__all__ = ["TIMonoPicture", "TIPicture", "TIImage",
           "L1", "RGBPalette", "RGB565"]



import os

from functools import total_ordering

from ..flags import *
from ..tokenizer.tokens.scripts import OsVersion, Tokens, TokenTrie
from ..tokenizer.tokens.scripts.parse import MODEL_ORDER


@total_ordering
class TIModel:
    """
    Data type for all model constants

    Every 83-series model is included in this module as a constant to use in code.
    Each model holds its (abbreviated) name, features, file magic, product ID, and native language.

    Models can also be used to obtain token maps and tries for tokenization and OS versions for compatibility checks.
    """

    MODELS = []
    """
    A list of all models
    """

    def __init__(self, name: str, features: 'TIFeature', magic: str, product_id: int, lang: str):
        self._name = name
        self._features = TIFeature(features)
        self._magic = magic
        self._product_id = product_id
        self._lang = lang

        with open(os.path.join(os.path.dirname(__file__), "../tokenizer/tokens/8X.xml"), encoding="UTF-8") as file:
            self._tokens = Tokens.from_xml_string(file.read(), self.OS("latest"))

        self._trie = {}
        for lang in self._tokens.langs:
            self._trie[lang] = TokenTrie.from_tokens(self._tokens, lang)

        self._trie[None] = self._trie["en"]

    def __eq__(self, other):
        return str(self) == str(other)

    def __ge__(self, other):
        return self.order >= other.order

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return self.name

    @property
    def features(self) -> 'TIFeature':
        """
        :return: This model's features
        """

        return self._features

    @property
    def lang(self) -> str:
        """
        :return: This model's native language
        """

        return self._lang

    @property
    def magic(self) -> str:
        """
        :return: This model's file magic
        """

        return self._magic

    @property
    def name(self) -> str:
        """
        :return: This model's (abbreviated) name
        """

        return self._name

    @property
    def order(self) -> int:
        """
        :return: This model's order within the chronology used by the token sheets
        """

        return MODEL_ORDER[self._name]

    @property
    def product_id(self) -> int:
        """
        :return: This model's product ID
        """

        return self._product_id

    @property
    def tokens(self) -> Tokens:
        """
        :return: The tokens supported by this model
        """

        return self._tokens

    def get_trie(self, lang: str = None) -> TokenTrie:
        """
        Gets the token trie for this model corresponding to a given language

        :param lang: A language code (defaults to English, `en`)
        :return: The token trie corresponding to `lang`
        """

        return self._trie[lang]

    def has(self, feature: 'TIFeature'):
        """
        Whether this model has a given feature

        :param feature: The feature to check
        :return: Whether this model has `feature`
        """

        return feature in self._features

    def OS(self, version: str = "") -> OsVersion:
        """
        An `OsVersion` with this model as its model and a supplied version

        :param version: An OS version number (defaults to the model's earliest OS)
        :return: An `OsVersion` for this model and `version`
        """

        return OsVersion(self.name, version)


class TIFeature(Flags):
    """
    Flags representing all calculator features
    """

    Complex = {0: 1}
    Flash = {1: 1}
    Apps = {2: 1}
    Clock = {3: 1}
    Color = {4: 1}
    ez80 = {5: 1}
    ExactMath = {6: 1}
    Python = {7: 1}


features82 = {}
features83 = features82 | TIFeature.Complex
features82a = features83 | TIFeature.Flash
features83p = features82a | TIFeature.Apps
features84p = features83p | TIFeature.Clock
features84pcse = features84p | TIFeature.Color
features84pce = features84pcse | TIFeature.ez80
features83pce = features84pce | TIFeature.ExactMath
features83pceep = features83pce | TIFeature.Python
features84pcepy = features84pce | TIFeature.Python
features82aep = features83pceep | {2: 0}

it = iter(MODEL_ORDER)
next(it)

TIModel.MODELS = [
    TI_82 := TIModel(next(it), features82, "**TI82**", 0x00, "en"),

    TI_83 := TIModel(next(it), features83, "**TI83**", 0x00, "en"),
    TI_82ST := TIModel(next(it), features83, "**TI83**", 0x00, "en"),
    TI_82ST_fr := TIModel(next(it), features83, "**TI83**", 0x00, "fr"),
    TI_76_fr := TIModel(next(it), features83, "**TI83**", 0x00, "fr"),

    TI_83P := TIModel(next(it), features83p, "**TI83F*", 0x04, "en"),
    TI_83PSE := TIModel(next(it), features83p, "**TI83F*", 0x04, "en"),
    TI_83P_fr := TIModel(next(it), features83p, "**TI83F*", 0x04, "fr"),
    TI_82P := TIModel(next(it), features83p, "**TI83F*", 0x04, "fr"),

    TI_84P := TIModel(next(it), features84p, "**TI83F*", 0x0A, "en"),
    TI_84PSE := TIModel(next(it), features84p, "**TI83F*", 0x0A, "en"),
    TI_83P_fr_USB := TIModel(next(it), features84p, "**TI83F*", 0x0A, "fr"),
    TI_84P_fr := TIModel(next(it), features84p, "**TI83F*", 0x0A, "fr"),
    TI_84PPSE := TIModel(next(it), features84p, "**TI83F*", 0x0A, "en"),

    TI_82A := TIModel(next(it), features82a, "**TI83F*", 0x0B, "fr"),
    TI_84PT := TIModel(next(it), features84p, "**TI83F*", 0x1B, "en"),

    TI_84PCSE := TIModel(next(it), features84pcse, "**TI83F*", 0x0F, "en"),

    TI_84PCE := TIModel(next(it), features84pce, "**TI83F*", 0x13, "en"),
    TI_84PCET := TIModel(next(it), features84pce, "**TI83F*", 0x13, "en"),
    TI_83PCE := TIModel(next(it), features83pce, "**TI83F*", 0x13, "fr"),
    TI_83PCEEP := TIModel(next(it), features83pceep, "**TI83F*", 0x13, "fr"),
    TI_84PCEPY := TIModel(next(it), features84pcepy, "**TI83F*", 0x13, "en"),
    TI_84PCETPE := TIModel(next(it), features84pcepy, "**TI83F*", 0x13, "en"),
    TI_82AEP := TIModel(next(it), features82aep, "**TI83F*", 0x00, "fr"),
]

__all__ = ["TI_82",
           "TI_83", "TI_82ST", "TI_82ST_fr", "TI_76_fr",
           "TI_83P", "TI_83PSE", "TI_83P_fr", "TI_82P",
           "TI_84P", "TI_84PSE", "TI_83P_fr_USB", "TI_84P_fr", "TI_84PPSE",
           "TI_84PT", "TI_82A",
           "TI_84PCSE",
           "TI_84PCE", "TI_84PCET", "TI_83PCE", "TI_83PCEEP", "TI_84PCEPY", "TI_84PCETPE", "TI_82AEP",
           "TIFeature", "TIModel", "OsVersion"]



import tivars.tokenizer.tokens.scripts.parse as parse

from .model import *


class OsVersions(parse.OsVersion):
    MATHPRINT = TI_84P.OS("2.53")

    ASMPRGM_DISABLED = TI_84PCE.OS("5.3.1")

    ASM_DISABLED_PCE = TI_83PCE.OS("5.5.1")
    ASM_DISABLED_CE = TI_84PCE.OS("5.6.0")


__all__ = ["OsVersions"]



from typing import Type
from warnings import warn

from tivars.models import *
from ..data import *
from ..var import TIEntry
from .numeric import *
from .real import *


class RealPart(Converter):
    """
    Converter for the real part of complex numbers

    Real parts are instances of `RealEntry`.
    """

    _T = RealEntry

    @classmethod
    def get(cls, data: bytes, *, instance=None) -> _T:
        """
        Converts `bytes` -> `RealType`

        :param data: The raw bytes to convert
        :param instance: The instance containing the data section
        :return: The real part of `data` converted to the appropriate type
        """

        return instance.real_type.get(data)

    @classmethod
    def set(cls, value: _T, **kwargs) -> bytes:
        """
        Converts `RealEntry` -> `bytes`

        :param value: The value to convert
        :return: The data of `value`
        """

        value.subtype_id = value.imag_subtype_id

        return type(value).set(value)


class ImaginaryPart(Converter):
    """
    Converter for the imaginary part of complex numbers

    Imaginary parts are instances of `RealEntry`.
    """

    _T = RealEntry

    @classmethod
    def get(cls, data: bytes, *, instance=None) -> _T:
        """
        Converts `bytes` -> `RealType`

        :param data: The raw bytes to convert
        :param instance: The instance containing the data section
        :return: The real part of `data` converted to the appropriate type
        """

        return instance.imag_type.get(data)

    @classmethod
    def set(cls, value: _T, *, instance=None, **kwargs) -> bytes:
        """
        Converts `RealEntry` -> `bytes`

        :param value: The value to convert
        :param instance: The instance containing the data section
        :return: The data of `value`
        """

        instance.imag_subtype_id = value.subtype_id = value.imag_subtype_id
        instance.coerce()

        return type(value).set(value)


class ComplexEntry(TIEntry):
    """
    Base class for complex numeric types

    This class handles floating-point types as well as the exact formats for the TI-83PCE and other newer models.
    The format for these types varies and is handled by `real_subtype_id` and `imag_subtype_id`.

    Two `RealEntry` types are used to form a single `ComplexEntry` corresponding to a complex number.
    These types need not be the same, and will have subtype IDs not corresponding to their native type IDs.

    The canonical type of the entire entry is determined by the imaginary part.
    The entry is coerced automatically if the imaginary part is updated.
    """

    versions = [0x00, 0x0B, 0x10]

    extensions = {
        None: "8xc",
        TI_83: "83c",
        TI_83P: "8xc"
    }

    min_data_length = 18

    is_exact = False
    """
    Whether this numeric type is exact
    """

    real_analogue = TIReal
    """
    The real type corresponding to this complex type used in an imaginary part
    """

    def __init__(self, init=None, *,
                 for_flash: bool = True, name: str = "A",
                 version: bytes = None, archived: bool = None,
                 data: bytearray = None):

        super().__init__(init, for_flash=for_flash, name=name, version=version, archived=archived, data=data)

    def __complex__(self):
        return self.complex()

    def __format__(self, format_spec: str) -> str:
        match format_spec:
            case "":
                return self.string()
            case "t":
                return squash(replacer(self.string(), {"i": "[i]", " 1[i]": "[i]", "+ 0[i]": "", "0 +": "",
                                                       "-": "~"}))
            case _:
                try:
                    return format(self.complex(), format_spec)

                except (TypeError, ValueError):
                    return super().__format__(format_spec)

    @Section(min_data_length)
    def calc_data(self) -> bytearray:
        """
        The data section of the entry

        Contains two real numbers, the real and imaginary parts
        """

    @View(calc_data, RealPart)[0:9]
    def real(self) -> RealEntry:
        """
        The real part of the complex number
        """

    @View(calc_data, ImaginaryPart)[9:18]
    def imag(self) -> RealEntry:
        """
        The imaginary part of the complex number
        """

    @View(calc_data, Bits[0:6])[0:1]
    def real_subtype_id(self) -> bytes:
        """
        The subtype ID of the real part
        """

    @View(calc_data, Bits[6:7])[0:1]
    def real_graph_bit(self) -> int:
        """
        Whether the entry is used during graphing
        """

    @View(calc_data, Bits[7:8])[0:1]
    def real_sign_bit(self) -> int:
        """
        The sign bit for the real part

        If this bit is set, the real part is negative.
        """

    @View(calc_data, Bits[0:6])[9:10]
    def imag_subtype_id(self) -> bytes:
        """
        The subtype ID of the complex part
        """

    @View(calc_data, Bits[6:7])[9:10]
    def imag_graph_bit(self) -> int:
        """
        Whether the entry is used during graphing
        """

    @View(calc_data, Bits[7:8])[9:10]
    def imag_sign_bit(self) -> int:
        """
        The sign bit for the complex part

        If this bit is set, the complex part is negative.
        """

    @property
    def real_type(self) -> Type['RealEntry']:
        """
        :return: The subclass of `RealEntry` corresponding to this entry's `real_subtype_id`.
        """

        return self.get_type(self.real_subtype_id).real_analogue

    @property
    def imag_type(self) -> Type['RealEntry']:
        """
        :return: The subclass of `RealEntry` corresponding to this entry's `imag_subtype_id`.
        """

        return self.get_type(self.imag_subtype_id).real_analogue

    def clear(self):
        super().clear()

        self.real_subtype_id = 0x0C
        self.imag_subtype_id = self.type_id

    def components(self) -> (RealEntry, RealEntry):
        """
        :return: The components of this complex number as a pair of `RealEntry` values
        """

        return self.real, self.imag

    def get_min_os(self, data: bytes = None) -> OsVersion:
        data = data or self.data

        match min(data[0], data[9]):
            case 0x0C:
                return TI_83.OS()

            case 0x1B:
                return TI_84P.OS("2.55")

            case _:
                return TI_83PCE()

    def get_version(self, data: bytes = None) -> int:
        data = data or self.data

        subtype_ids = data[0], data[9]
        if subtype_ids == (0x0C, 0x0C):
            return 0x00

        elif 0x1B in subtype_ids:
            return 0x0B

        else:
            return 0x10

    def supported_by(self, model: TIModel) -> bool:
        return super().supported_by(model) and (self.get_version() <= 0x0B or model.has(TIFeature.ExactMath))

    @Loader[complex, float, int]
    def load_complex(self, comp: complex):
        """
        Loads this complex number from a `complex`, upcasting as necessary

        :param comp: The complex number to load
        """

        raise NotImplementedError

    def complex(self):
        """
        :return: The `complex` corresponding to this complex number
        """

        return self.real.float() + 1j * self.imag.float()

    @Loader[str]
    def load_string(self, string: str):
        string = replacer(squash(string), {"*": "", "-": "+-", "[i]": "i", "j": "i"})

        # Split into real and imaginary components
        parts = string.strip("+").split("+")
        if not parts[0]:
            parts = parts[1:]

        # Add missing components
        if len(parts) == 1:
            if "i" in parts[0]:
                parts = ["0", parts[0]]

            else:
                parts = [parts[0], "0i"]

        parts[1] = parts[1].replace("i", "") or "1"

        try:
            self.real = self.real_type(parts[0])

        except (TypeError, ValueError):
            for type_id in [0x00, 0x18, 0x1C, 0x20, 0x21]:
                if type_id == self.real_subtype_id:
                    continue

                try:
                    self.real = self.get_type(type_id)(parts[0])
                    break

                except (TypeError, ValueError):
                    continue

            else:
                raise ValueError(f"could not parse real part '{parts[0]}'")

        self.imag = self.imag_type(parts[1])

    def string(self) -> str:
        def make_imag(entry: 'RealEntry') -> str:
            match entry.type_id:
                case 0x18 | 0x21:
                    return str(entry).replace(" /", "i /")

                case 0x1C:
                    return f"{entry} * i"

                case _:
                    return f"{entry}i"

        match str(self.real), str(self.imag):
            case "0", "0":
                return "0"

            case _, "0":
                return str(self.real)

            case "0", _:
                return make_imag(self.imag)

            case _:
                return replacer(f"{self.real} + " + make_imag(self.imag), {"+ -": "- ", " 1i": " i"})

    def coerce(self):
        self.type_id = self.imag_subtype_id
        if subclass := self.get_type(self.type_id):
            self.__class__ = subclass

        else:
            warn(f"Subtype ID 0x{self.type_id:02x} is not recognized; entry will not be coerced to its proper type.",
                 BytesWarning)


class TIComplex(ComplexEntry, register=True):
    """
    Parser for complex entries with floating point imaginary part

    A `TIComplex` has a `TIReal` as its imaginary part.
    """

    real_analogue = TIReal

    _type_id = 0x0C

    @Loader[complex, float, int]
    def load_complex(self, comp: complex):
        real, imag = self.real_type(), self.imag_type()
        comp = complex(comp)

        real.load_float(comp.real)
        imag.load_float(comp.imag)

        self.real, self.imag = real, imag


class TIComplexFraction(TIComplex, register=True):
    """
    Parser for complex entries with fractional imaginary part

    A `TIComplexFraction` has a `TIRealFraction` as its imaginary part.
    """

    is_exact = True

    real_analogue = TIRealFraction

    _type_id = 0x1B


class TIComplexRadical(ComplexEntry, register=True):
    r"""
    Parser for complex entries with radical imaginary part

    A `TIComplexRadical` has a `TIRealRadical` as its imaginary part.
    """

    flash_only = True

    is_exact = True

    real_analogue = TIRealRadical

    _type_id = 0x1D

    @Loader[complex, float, int]
    def load_complex(self, comp: complex):
        return NotImplemented


class TIComplexPi(TIComplex, register=True):
    """
    Parser for complex entries with imaginary part an integer multiple of π

    A `TIComplexPi` has a `TIRealPi` as its imaginary part.
    """

    flash_only = True

    is_exact = True

    real_analogue = TIRealPi

    _type_id = 0x1E

    @Loader[complex, float, int]
    def load_complex(self, comp: complex):
        return NotImplemented


class TIComplexPiFraction(TIComplexPi, TIComplexFraction, register=True):
    """
    Parser for complex entries with imaginary part a fractional multiple of π

    A `TIComplexPiFraction` has a `TIRealPiFraction` as its imaginary part.
    """

    flash_only = True

    is_exact = True

    real_analogue = TIRealPiFraction

    _type_id = 0x1F



from .model import *
from .versions import *

__all__ = ["TI_82",
           "TI_83", "TI_82ST", "TI_82ST_fr", "TI_76_fr",
           "TI_83P", "TI_83PSE", "TI_83P_fr", "TI_82P",
           "TI_84P", "TI_84PSE", "TI_83P_fr_USB", "TI_84P_fr", "TI_84PPSE",
           "TI_84PT", "TI_82A",
           "TI_84PCSE",
           "TI_84PCE", "TI_84PCET", "TI_83PCE", "TI_83PCEEP", "TI_84PCEPY", "TI_84PCETPE", "TI_82AEP",
           "TIFeature", "TIModel", "OsVersion", "OsVersions"]



from .appvar import *
from .complex import *
from .gdb import *
from .list import *
from .matrix import *
from .real import *
from .picture import *
from .settings import *
from .tokenized import *

from ..var import TIEntry


__all__ = ["TIEntry",
           "TIReal",
           "TIRealList", "TIMatrix",
           "TIEquation", "TIString", "TIProgram", "TIProtectedProgram",
           "TIPicture", "TIMonoPicture",
           "TIMonoGDB", "TIGDB", "TIGraphedEquation",
           "TIMonoFuncGDB", "TIMonoParamGDB", "TIMonoPolarGDB", "TIMonoSeqGDB",
           "TIFuncGDB", "TIParamGDB", "TIPolarGDB", "TISeqGDB",
           "EquationFlags", "GraphMode", "GraphStyle", "GraphColor", "GlobalLineStyle",
           "TIComplex", "TIComplexList",
           "TIUndefinedReal",
           "TIWindowSettings", "TIRecallWindow", "TITableSettings",
           "TIAppVar",
           "TIRealFraction",
           "TIImage",
           "TIComplexFraction",
           "TIRealRadical", "TIComplexRadical",
           "TIComplexPi", "TIComplexPiFraction",
           "TIRealPi", "TIRealPiFraction"
           ]



import json

from io import BytesIO
from typing import ByteString
from warnings import warn

from tivars.models import *
from ..data import *
from ..var import TIEntry
from .real import GraphRealEntry


class SettingsEntry(TIEntry):
    """
    Base class for settings entries

    A settings entry stores all parameters for the different plot windows or tables
    """

    min_data_length = 2

    def __init__(self, init=None, *,
                 for_flash: bool = True, name: str = "Window",
                 version: bytes = None, archived: bool = None,
                 data: bytearray = None):
        super().__init__(init, for_flash=for_flash, name=name, version=version, archived=archived, data=data)

    @Loader[ByteString, BytesIO]
    def load_bytes(self, data: bytes | BytesIO):
        super().load_bytes(data)

        if self.calc_data[:len(self.leading_bytes)] != self.leading_bytes:
            warn(f"The entry has unexpected leading bytes "
                 f"(expected {self.leading_bytes}, got {self.calc_data[:len(self.leading_bytes)]}).",
                 BytesWarning)

    @Loader[dict]
    def load_dict(self, dct: dict):
        """
        Loads a JSON `dict` into this settings entry

        :param dct: The dict to load
        """

        for var, value in dct:
            if not hasattr(self, var):
                warn(f"Unrecognized window setting ({var}).",
                     UserWarning)
            else:
                setattr(self, var, GraphRealEntry(value))

    def dict(self) -> dict:
        """
        :return: A `dict` representing this settings entry in JSON format
        """

        raise NotImplementedError

    @Loader[str]
    def load_string(self, string: str):
        self.load_dict(json.loads(string))

    def string(self) -> str:
        return json.dumps(self.dict())


class TIWindowSettings(SettingsEntry, register=True):
    """
    Parser for window settings

    A `TIWindowSettings` stores all plot window parameters as a contiguous stream of `TIReal` values.
    """

    extensions = {
        None: "8xw",
        TI_82: "82w",
        TI_83: "83w",
        TI_83P: "8xw"
    }

    min_data_length = 210

    leading_bytes = b'\xD0\x00\x00'

    _type_id = 0x0F

    def __init__(self, init=None, *,
                 for_flash: bool = True, name: str = "Window",
                 version: bytes = None, archived: bool = None,
                 data: bytearray = None):

        super().__init__(init, for_flash=for_flash, name=name, version=version, archived=archived, data=data)

    @Section(8, String)
    def name(self) -> str:
        pass

    @Section(min_data_length)
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, Bytes)[0:3]
    def table_magic(self) -> bytearray:
        """
        Magic identifying the file as window settings

        This value is always `0xD00000`.
        """

    @View(calc_data, Bytes)[3:]
    def data(self) -> bytearray:
        pass

    @View(calc_data, GraphRealEntry)[3:12]
    def Xmin(self) -> GraphRealEntry:
        """
        Xmin: the X-coordinate of the left edge of the graphscreen
        """

    @View(calc_data, GraphRealEntry)[12:21]
    def Xmax(self) -> GraphRealEntry:
        """
        Xmax: the X-coordinate of the right edge of the graphscreen
        """

    @View(calc_data, GraphRealEntry)[21:30]
    def Xscl(self) -> GraphRealEntry:
        """
        Xscl: the separation between ticks on the X-axis
        """

    @View(calc_data, GraphRealEntry)[30:39]
    def Ymin(self) -> GraphRealEntry:
        """
        Ymin: the Y-coordinate of the bottom edge of the graphscreen
        """

    @View(calc_data, GraphRealEntry)[39:48]
    def Ymax(self) -> GraphRealEntry:
        """
        Ymax: the Y-coordinate of the top edge of the graphscreen
        """

    @View(calc_data, GraphRealEntry)[48:57]
    def Yscl(self) -> GraphRealEntry:
        """
        Yscl: the separation between ticks on the Y-axis
        """

    @View(calc_data, GraphRealEntry)[57:66]
    def Thetamin(self) -> GraphRealEntry:
        """
        Θmin: the initial angle for polar plots
        """

    @View(calc_data, GraphRealEntry)[66:75]
    def Thetamax(self) -> GraphRealEntry:
        """
        Θmax: the final angle for polar plots
        """

    @View(calc_data, GraphRealEntry)[75:84]
    def Thetastep(self) -> GraphRealEntry:
        """
        Θstep: the angle increment for polar plots
        """

    @View(calc_data, GraphRealEntry)[84:93]
    def Tmin(self) -> GraphRealEntry:
        """
        Tmin: the initial time for parametric plots
        """

    @View(calc_data, GraphRealEntry)[93:102]
    def Tmax(self) -> GraphRealEntry:
        """
        Tmax: the final time for parametric plots
        """

    @View(calc_data, GraphRealEntry)[102:111]
    def Tstep(self) -> GraphRealEntry:
        """
        Tstep: the time increment for parametric plots
        """

    @View(calc_data, GraphRealEntry)[111:120]
    def PlotStart(self, value) -> GraphRealEntry:
        """
        PlotStart: the initial value of 𝑛 for sequential plots

        The value must be an integer.
        """

        if int(value) != float(value):
            warn(f"Expected an integer, got {float(value)}.",
                 UserWarning)

        return value

    @View(calc_data, GraphRealEntry)[120:129]
    def nMax(self, value) -> GraphRealEntry:
        """
        𝑛Max: the final value of 𝑛 for sequential equations and plots

        The value must be an integer.
        """

        if int(value) != float(value):
            warn(f"Expected an integer, got {float(value)}.",
                 UserWarning)

        return value

    @View(calc_data, GraphRealEntry)[129:138]
    def unMin0(self) -> GraphRealEntry:
        """
        u(𝑛Min): the initial value of u at 𝑛Min
        """

    @View(calc_data, GraphRealEntry)[138:147]
    def vnMin0(self) -> GraphRealEntry:
        """
        v(𝑛Min): the initial value of v at 𝑛Min
        """

    @View(calc_data, GraphRealEntry)[147:156]
    def nMin(self, value) -> GraphRealEntry:
        """
        𝑛Min: the initial value of 𝑛 for sequential plots

        The value must be an integer.
        """

        if int(value) != float(value):
            warn(f"Expected an integer, got {float(value)}.",
                 UserWarning)

        return value

    @View(calc_data, GraphRealEntry)[156:165]
    def unMin1(self) -> GraphRealEntry:
        """
        u(𝑛Min+1): the initial value of u at 𝑛Min + 1
        """

    @View(calc_data, GraphRealEntry)[165:174]
    def vnMin1(self) -> GraphRealEntry:
        """
        v(𝑛Min+1): the initial value of v at 𝑛Min + 1
        """

    @View(calc_data, GraphRealEntry)[174:183]
    def wnMin0(self) -> GraphRealEntry:
        """
        w(𝑛Min): the initial value of w at 𝑛Min
        """

    @View(calc_data, GraphRealEntry)[183:192]
    def PlotStep(self, value) -> GraphRealEntry:
        """
        PlotStep: the 𝑛 increment for sequential plots

        The value must be an integer.
        """

        if int(value) != float(value):
            warn(f"Expected an integer, got {float(value)}.",
                 UserWarning)

        return value

    @View(calc_data, GraphRealEntry)[192:201]
    def Xres(self, value) -> GraphRealEntry:
        """
        Xres: the pixel separation of points in a function plot

        The value must be an integer in [1,8]
        """

        if int(value) != float(value) or not 1 <= int(value) <= 8:
            warn(f"Expected an integer between 1 and 8, got {float(value)}.",
                 UserWarning)

        return value

    @View(calc_data, GraphRealEntry)[201:210]
    def wnMin1(self) -> GraphRealEntry:
        """
        w(𝑛Min+1): the initial value of w at 𝑛Min + 1
        """

    def dict(self) -> dict:
        return {
            "Xmin": self.Xmin.json_number(),
            "Xmax": self.Xmax.json_number(),
            "Xscl": self.Xscl.json_number(),
            "Ymin": self.Ymin.json_number(),
            "Ymax": self.Ymax.json_number(),
            "Yscl": self.Yscl.json_number(),
            "Thetamin": self.Thetamin.json_number(),
            "Thetamax": self.Thetamax.json_number(),
            "Thetastep": self.Thetastep.json_number(),
            "Tmin": self.Tmin.json_number(),
            "Tmax": self.Tmax.json_number(),
            "Tstep": self.Tstep.json_number(),
            "PlotStart": int(self.PlotStart),
            "nMax": int(self.nMax),
            "unMin0": self.unMin0.json_number(),
            "vnMin0": self.vnMin0.json_number(),
            "nMin": int(self.nMin),
            "unMin1": self.unMin1.json_number(),
            "vnMin1": self.vnMin1.json_number(),
            "wnMin0": self.wnMin0.json_number(),
            "PlotStep": int(self.PlotStep),
            "Xres": int(self.Xres),
            "wnMin1": self.wnMin1.json_number()
        }


class TIRecallWindow(SettingsEntry, register=True):
    """
    Parser for recalled windows

    A `TIRecallWindow` stores all plot window parameters as a contiguous stream of `TIReal` values.
    """

    extensions = {
        None: "8xz",
        TI_82: "82z",
        TI_83: "83z",
        TI_83P: "8xz"
    }

    min_data_length = 209

    leading_bytes = b'\xCF\x00'

    _type_id = 0x10

    def __init__(self, init=None, *,
                 for_flash: bool = True, name: str = "RclWindw",
                 version: bytes = None, archived: bool = None,
                 data: bytearray = None):

        super().__init__(init, for_flash=for_flash, name=name, version=version, archived=archived, data=data)

    @Section(8, String)
    def name(self) -> str:
        """
        The name of the entry

        Always equal to RclWindw.
        """

    @Section(min_data_length)
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, Bytes)[0:2]
    def recall_magic(self) -> bytearray:
        """
        Magic identifying the file as a recalled window

        This value is always `0xCF00`.
        """

    @View(calc_data, Bytes)[2:]
    def data(self) -> bytearray:
        pass

    @View(calc_data, GraphRealEntry)[2:11]
    def Xmin(self) -> GraphRealEntry:
        """
        Xmin: the X-coordinate of the left edge of the graphscreen
        """

    @View(calc_data, GraphRealEntry)[11:20]
    def Xmax(self) -> GraphRealEntry:
        """
        Xmax: the X-coordinate of the right edge of the graphscreen
        """

    @View(calc_data, GraphRealEntry)[20:29]
    def Xscl(self) -> GraphRealEntry:
        """
        Xscl: the separation between ticks on the X-axis
        """

    @View(calc_data, GraphRealEntry)[29:38]
    def Ymin(self) -> GraphRealEntry:
        """
        Ymin: the Y-coordinate of the bottom edge of the graphscreen
        """

    @View(calc_data, GraphRealEntry)[38:47]
    def Ymax(self) -> GraphRealEntry:
        """
        Ymax: the Y-coordinate of the top edge of the graphscreen
        """

    @View(calc_data, GraphRealEntry)[47:56]
    def Yscl(self) -> GraphRealEntry:
        """
        Yscl: the separation between ticks on the Y-axis
        """

    @View(calc_data, GraphRealEntry)[56:65]
    def Thetamin(self) -> GraphRealEntry:
        """
        Θmin: the initial angle for polar plots
        """

    @View(calc_data, GraphRealEntry)[65:74]
    def Thetamax(self) -> GraphRealEntry:
        """
        Θmax: the final angle for polar plots
        """

    @View(calc_data, GraphRealEntry)[74:83]
    def Thetastep(self) -> GraphRealEntry:
        """
        Θstep: the angle increment for polar plots
        """

    @View(calc_data, GraphRealEntry)[83:92]
    def Tmin(self) -> GraphRealEntry:
        """
        Tmin: the initial time for parametric plots
        """

    @View(calc_data, GraphRealEntry)[92:101]
    def Tmax(self) -> GraphRealEntry:
        """
        Tmax: the final time for parametric plots
        """

    @View(calc_data, GraphRealEntry)[101:110]
    def Tstep(self) -> GraphRealEntry:
        """
        Tstep: the time increment for parametric plots
        """

    @View(calc_data, GraphRealEntry)[110:119]
    def PlotStart(self, value) -> GraphRealEntry:
        """
        PlotStart: the initial value of 𝑛 for sequential plots

        The value must be an integer.
        """

        if int(value) != float(value):
            warn(f"Expected an integer, got {float(value)}.",
                 UserWarning)

        return value

    @View(calc_data, GraphRealEntry)[119:128]
    def nMax(self, value) -> GraphRealEntry:
        """
        𝑛Max: the final value of 𝑛 for sequential equations and plots

        The value must be an integer.
        """

        if int(value) != float(value):
            warn(f"Expected an integer, got {float(value)}.",
                 UserWarning)

        return value

    @View(calc_data, GraphRealEntry)[128:137]
    def unMin0(self) -> GraphRealEntry:
        """
        u(𝑛Min): the initial value of u at 𝑛Min
        """

    @View(calc_data, GraphRealEntry)[137:146]
    def vnMin0(self) -> GraphRealEntry:
        """
        v(𝑛Min): the initial value of v at 𝑛Min
        """

    @View(calc_data, GraphRealEntry)[146:155]
    def nMin(self, value) -> GraphRealEntry:
        """
        𝑛Min: the initial value of 𝑛 for sequential equations

        The value must be an integer.
        """

        if int(value) != float(value):
            warn(f"Expected an integer, got {float(value)}.",
                 UserWarning)

        return value

    @View(calc_data, GraphRealEntry)[155:164]
    def unMin1(self) -> GraphRealEntry:
        """
        u(𝑛Min + 1): the initial value of u at 𝑛Min + 1
        """

    @View(calc_data, GraphRealEntry)[164:173]
    def vnMin1(self) -> GraphRealEntry:
        """
        v(𝑛Min + 1): the initial value of v at 𝑛Min + 1
        """

    @View(calc_data, GraphRealEntry)[173:182]
    def wnMin0(self) -> GraphRealEntry:
        """
        w(𝑛Min): the initial value of w at 𝑛Min
        """

    @View(calc_data, GraphRealEntry)[182:191]
    def PlotStep(self, value) -> GraphRealEntry:
        """
        PlotStep: the 𝑛 increment for sequential plots

        The value must be an integer.
        """

        if int(value) != float(value):
            warn(f"Expected an integer, got {float(value)}.",
                 UserWarning)

        return value

    @View(calc_data, GraphRealEntry)[191:200]
    def Xres(self, value) -> GraphRealEntry:
        """
        Xres: the pixel separation of points in a function plot

        The value must be an integer in [1,8]
        """

        if int(value) != float(value) or not 1 <= int(value) <= 8:
            warn(f"Expected an integer between 1 and 8, got {float(value)}.",
                 UserWarning)

        return value

    @View(calc_data, GraphRealEntry)[200:209]
    def wnMin1(self) -> GraphRealEntry:
        """
        w(𝑛Min + 1): the initial value of w at 𝑛Min + 1
        """

    def dict(self) -> dict:
        return {
            "Xmin": self.Xmin.json_number(),
            "Xmax": self.Xmax.json_number(),
            "Xscl": self.Xscl.json_number(),
            "Ymin": self.Ymin.json_number(),
            "Ymax": self.Ymax.json_number(),
            "Yscl": self.Yscl.json_number(),
            "Thetamin": self.Thetamin.json_number(),
            "Thetamax": self.Thetamax.json_number(),
            "Thetastep": self.Thetastep.json_number(),
            "Tmin": self.Tmin.json_number(),
            "Tmax": self.Tmax.json_number(),
            "Tstep": self.Tstep.json_number(),
            "PlotStart": int(self.PlotStart),
            "nMax": int(self.nMax),
            "unMin0": self.unMin0.json_number(),
            "vnMin0": self.vnMin0.json_number(),
            "nMin": int(self.nMin),
            "unMin1": self.unMin1.json_number(),
            "vnMin1": self.vnMin1.json_number(),
            "wnMin0": self.wnMin0.json_number(),
            "PlotStep": int(self.PlotStep),
            "Xres": int(self.Xres),
            "wnMin1": self.wnMin1.json_number()
        }


class TITableSettings(SettingsEntry, register=True):
    """
    Parser for table settings

    A `TITableSettings` stores all plot table parameters as a contiguous stream of `TIReal` values.
    """

    extensions = {
        None: "8xt",
        TI_82: "82t",
        TI_83: "83t",
        TI_83P: "8xt"
    }

    min_data_length = 20

    leading_bytes = b'\x12\x00'

    _type_id = 0x11

    def __init__(self, init=None, *,
                 for_flash: bool = True, name: str = "TblSet",
                 version: bytes = None, archived: bool = None,
                 data: bytearray = None):

        super().__init__(init, for_flash=for_flash, name=name, version=version, archived=archived, data=data)

    @Section(8, String)
    def name(self) -> str:
        """
        The name of the entry

        Always equal to TblSet.
        """

    @Section(min_data_length)
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, Bytes)[0:2]
    def table_magic(self) -> bytearray:
        """
        Magic identifying the file as an image

        This value is always `0x1200`.
        """

    @View(calc_data, Bytes)[2:]
    def data(self) -> bytearray:
        pass

    @View(calc_data, GraphRealEntry)[2:11]
    def TblMin(self, value) -> GraphRealEntry:
        """
        TblMin: the initial value for the table

        The value must be an integer.
        """

        if int(value) != float(value):
            warn(f"Expected an integer for TblMin, got {float(value)}.",
                 UserWarning)

        return value

    @View(calc_data, GraphRealEntry)[11:20]
    def DeltaTbl(self, value) -> GraphRealEntry:
        """
        ΔTbl: the increment for the table

        The value must be an integer.
        """

        if int(value) != float(value):
            warn(f"Expected an integer for ΔTbl, got {float(value)}.",
                 UserWarning)

        return value

    def dict(self) -> dict:
        return {
            "TblMin": int(self.TblMin),
            "DeltaTbl": int(self.DeltaTbl)
        }


__all__ = ["TIWindowSettings", "TIRecallWindow", "TITableSettings"]



from io import BytesIO
from typing import ByteString, Iterator
from warnings import warn

from tivars.models import *
from ..data import *
from ..var import TIEntry
from .real import RealEntry


class TIMatrix(TIEntry, register=True):
    """
    Parser for the matrix type

    A `TIMatrix` is a two-dimensional array of `RealEntry` elements.
    Exact types are supported, but complex numbers are not.
    """

    versions = [0x10, 0x0B, 0x00]

    extensions = {
        None: "8xm",
        TI_82: "82m",
        TI_83: "83m",
        TI_83P: "8xm"
    }

    min_data_length = 2

    _type_id = 0x02

    def __init__(self, init=None, *,
                 for_flash: bool = True, name: str = "[A]",
                 version: bytes = None, archived: bool = None,
                 data: ByteString = None):

        super().__init__(init, for_flash=for_flash, name=name, version=version, archived=archived, data=data)

    def __format__(self, format_spec: str) -> str:
        match format_spec:
            case "":
                inner_sep, outer_sep = ", ", ", "
            case "t":
                inner_sep, outer_sep = ",", ""
            case _:
                return super().__format__(format_spec)

        return "[" + outer_sep.join(f"[{inner_sep.join(format(entry, format_spec)for entry in row)}]"
                                    for row in self.matrix()) + "]"

    def __iter__(self) -> Iterator[RealEntry]:
        for row in self.matrix():
            for entry in row:
                yield entry

    @Section()
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, Integer)[0:1]
    def width(self, value: int) -> int:
        """
        The number of columns in the matrix

        TI-OS imposes a limit of 99.
        Additionally, TI-OS imposes a limit of 400 total elements.
        """

        if value > 99:
            warn(f"The matrix is too wide ({value} > 99).",
                 UserWarning)

        if value * self.height > 400:
            warn(f"The matrix is too big ({value * self.height} > 400).",
                 UserWarning)

        return value

    @View(calc_data, Integer)[1:2]
    def height(self, value: int) -> int:
        """
        The number of rows in the matrix

        TI-OS imposes a limit of 99.
        Additionally, TI-OS imposes a limit of 400 total elements.
        """

        if value > 99:
            warn(f"The matrix is too tall ({value} > 99).",
                 UserWarning)

        if self.width * value > 400:
            warn(f"The matrix is too big ({self.width * value} > 400).",
                 UserWarning)

        return value

    @View(calc_data, Bytes)[2:]
    def data(self) -> bytearray:
        pass

    @property
    def size(self) -> int:
        """
        :return: The number of elements in the matrix
        """

        return self.width * self.height

    def get_min_os(self, data: bytes = None) -> OsVersion:
        it = zip(*[iter(data or self.data)] * RealEntry.min_data_length)
        return max(map(RealEntry().get_min_os, it), default=OsVersions.INITIAL)

    def get_version(self, data: bytes = None) -> int:
        it = zip(*[iter(data or self.data)] * RealEntry.min_data_length)
        version = max(map(RealEntry().get_version, it), default=0x00)

        if version > 0x1B:
            return 0x10

        elif version == 0x1B:
            return 0x0B

        else:
            return 0x00

    def supported_by(self, model: TIModel) -> bool:
        return super().supported_by(model) and (self.get_version() <= 0x0B or model.has(TIFeature.ExactMath))

    @Loader[ByteString, BytesIO]
    def load_bytes(self, data: bytes | BytesIO):
        super().load_bytes(data)

        if self.data_length // RealEntry.min_data_length != self.size:
            warn(f"The matrix has an unexpected size "
                 f"(expected {self.data_length // RealEntry.min_data_length}, got {self.size}).",
                 BytesWarning)

    def load_data_section(self, data: BytesIO):
        width = int.from_bytes(width_byte := data.read(1), 'little')
        height = int.from_bytes(height_byte := data.read(1), 'little')
        self.raw.calc_data = bytearray(width_byte + height_byte + data.read(width * height))

    @Loader[list]
    def load_matrix(self, matrix: list[list[RealEntry]]):
        """
        Loads a two-dimensional `list` into this matrix

        :param matrix: The matrix to load
        """

        if len({len(row) for row in matrix}) != 1:
            raise IndexError("matrix has uneven rows")

        self.load_bytes(bytes([len(matrix[0])]) + bytes([len(matrix)]) +
                        b''.join(entry.calc_data for row in matrix for entry in row))

    def matrix(self) -> list[list[RealEntry]]:
        """
        :return: A two-dimensional `list` of the elements in this matrix
        """

        it = zip(*[iter(self.data)] * RealEntry.min_data_length)
        return [[RealEntry(for_flash=self.meta_length > TIEntry.base_meta_length, data=bytes(data))
                 for data in row] for row in zip(*[it] * self.width)]

    @Loader[str]
    def load_string(self, string: str):
        matrix = []

        for string in ''.join(string.split())[1:-1].replace("],[", "][").split("]["):
            row = []
            for item in string.replace("[", "").replace("]", "").split(","):
                entry = RealEntry()
                entry.load_string(item)
                row.append(entry)

            matrix.append(row.copy())

        self.load_matrix(matrix)

    def string(self) -> str:
        return format(self, "")


__all__ = ["TIMatrix"]



import copy
import decimal as dec
import fractions as frac
import re

from typing import Type
from warnings import warn

from tivars.models import *
from ..data import *
from ..var import TIEntry
from .numeric import *


class RealEntry(TIEntry):
    """
    Base class for real numeric types

    This class handles floating-point types as well as the exact formats for the TI-83PCE and other newer models.
    The format for these types varies and is handled by the `subtype_id`, the first six bits of the first data byte.

    Two `RealEntry` types are used to form a single `ComplexEntry` corresponding to a complex number.
    """

    _T = 'RealEntry'

    extensions = {
        None: "8xn",
        TI_82: "82n",
        TI_83: "83n",
        TI_83P: "8xn"
    }

    min_data_length = 9

    min_exponent = 0x00
    """
    The smallest allowed floating point exponent
    """

    is_exact = False
    """
    Whether this numeric type is exact
    """

    imag_subtype_id = None
    """
    The subtype ID this type receives if used as an imaginary part
    """

    def __init__(self, init=None, *,
                 for_flash: bool = True, name: str = "A",
                 version: bytes = None, archived: bool = None,
                 data: bytes = None):

        super().__init__(init, for_flash=for_flash, name=name, version=version, archived=archived, data=data)

    def __float__(self) -> float:
        return self.float()

    def __format__(self, format_spec: str) -> str:
        match format_spec:
            case "":
                return self.string()
            case "j":
                return str(self.json_number())
            case "t":
                return squash(re.sub(r"(√\d*)", r"sqrt(\1)", replacer(self.string(), {"-": "~", "/": "n/d"})))
            case _:
                try:
                    return format(self.decimal(), format_spec)

                except (TypeError, ValueError):
                    return super().__format__(format_spec)

    def __int__(self) -> int:
        return int(self.float())

    def __neg__(self) -> 'RealEntry':
        negated = copy.copy(self)
        negated.sign_bit = 1 - negated.sign_bit

        return negated

    @Section(min_data_length)
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, Bits[0:6])[0:1]
    def subtype_id(self) -> bytes:
        """
        The subtype ID of the number

        Differentiates the exact radical and fraction formats.
        """

    @View(calc_data, Bits[6:7])[0:1]
    def graph_bit(self) -> int:
        """
        Whether the entry is used during graphing
        """

    @View(calc_data, Bits[7:8])[0:1]
    def sign_bit(self) -> int:
        """
        The sign bit for the number

        If this bit is used, the number is negative when set.
        """

    @property
    def subtype(self) -> Type['RealEntry']:
        """
        :return: The subtype of this real number
        """

        return self.get_type(self.subtype_id)

    @property
    def sign(self) -> int:
        """
        :return: The sign of this real number
        """

        return -1 if self.sign_bit else 1

    def clear(self):
        super().clear()

        if self._type_id is not None:
            self.subtype_id = self._type_id

    def get_min_os(self, data: bytes = None) -> OsVersion:
        data = data or self.data

        match data[0]:
            case 0x18 | 0x19:
                return TI_84P.OS("2.53")

            case _:
                return OsVersions.INITIAL

    def supported_by(self, model: TIModel) -> bool:
        return super().supported_by(model) and (self.subtype_id <= 0x19 or model.has(TIFeature.ExactMath))

    @Loader[dec.Decimal]
    def load_decimal(self, decimal: dec.Decimal):
        """
        Loads a `dec.Decimal` into this real number

        :param decimal: The decimal to load
        """

        raise NotImplementedError

    def decimal(self) -> dec.Decimal:
        """
        :return: A `dec.Decimal` object corresponding to this real number
        """

        raise NotImplementedError

    @Loader[float, int]
    def load_float(self, decimal: float):
        """
        Loads a `float` into this real number, upcasting as necessary

        :param decimal: The float to load
        """

        self.load_decimal(dec.Decimal(decimal))

    def json_number(self) -> float | str:
        """
        Encoder for JSON implementations with potentially low precision

        :return: A `float` or `str` depending on whether this real number can be contained in a single-precision float
        """

        if len(str(number := self.float())) <= 6:
            return number

        else:
            return str(number)

    def float(self) -> float:
        """
        :return: The `float` corresponding to this real number
        """

        return float(self.decimal())

    def coerce(self):
        if self._type_id is None:
            self.type_id = self.subtype_id

            super().coerce()


class GraphRealEntry(RealEntry):
    """
    Warning converter for real numeric types supported by graph parameters

    Values used for plotting (e.g. Xmin) will behave unexpectedly if set to a radical or π type.
    """

    _T = 'RealEntry'

    @classmethod
    def set(cls, value: _T, **kwargs) -> bytes:
        if type(value) not in (TIReal, TIUndefinedReal, TIRealFraction):
            warn(f"Graph parameters cannot store {type(value)} values correctly.",
                 UserWarning)

        return super().set(value)


class TIReal(RealEntry, register=True):
    """
    Parser for real floating point values

    A `TIReal` has 8 exponent bits and 14 decimal mantissa digits.

    A `TIReal` entry can be used to form `TIComplex` or `TIComplexPi` complex numbers.
    """

    min_data_length = 9

    imag_subtype_id = 0x0C

    _type_id = 0x00

    @Section(min_data_length)
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, Integer)[1:2]
    def exponent(self) -> int:
        """
        The exponent of the real number

        The exponent has a bias of 0x80.
        """

    @View(calc_data, BCD)[2:9]
    def mantissa(self) -> int:
        """
        The mantissa of the real number

        The mantissa is 14 digits stored in BCD format, two digits per byte.
        """

    @Loader[dec.Decimal]
    def load_decimal(self, decimal: dec.Decimal):
        self.load_string(str(decimal))

    def decimal(self) -> dec.Decimal:
        with dec.localcontext() as ctx:
            ctx.prec = 14
            decimal = dec.Decimal(self.sign * self.mantissa)
            decimal *= dec.Decimal(10) ** (self.exponent - 0x80 - 13)

        return decimal

    @Loader[str]
    def load_string(self, string: str):
        if not string:
            self.mantissa, self.exponent, self.sign_bit = 0, 0x80, False
            return

        # Normalize string
        string = replacer(squash(string).lower(), {"~": "-", "|e": "e"})

        if "e" not in string:
            string += "e0"

        if "." not in string:
            string = string.replace("e", ".e")

        neg = string.startswith("-")
        string = string.strip("+-")

        # Obtain integer and decimal parts
        number, exponent = string.split("e")
        integer, decimal = number.split(".")
        integer, decimal = integer or "0", decimal or "0"

        if int(integer) == int(decimal) == 0:
            self.mantissa, self.exponent, self.sign_bit = 0, 0x80, neg
            return

        # Adjust exponent to make integer mantissa
        exponent = int(exponent or "0")
        while not 0 < (value := int(integer)) < 10:
            if value == 0:
                integer, decimal = decimal[0], decimal[1:]
                exponent -= 1

            else:
                integer, decimal = integer[:-1], integer[-1] + decimal
                exponent += 1

        self.mantissa, self.exponent, self.sign_bit = int((integer + decimal).ljust(14, "0")[:14]), exponent + 0x80, neg

    def string(self) -> str:
        string = f"{self.decimal():.14g}".rstrip("0").rstrip(".")

        if string.startswith("0e"):
            return "0"
        else:
            return string


class TIUndefinedReal(TIReal, register=True):
    """
    Parser for undefined real values

    A `TIUndefinedReal` is precisely a `TIReal` but marked as undefined for use in initial sequence values
    """

    _T = 'TIUndefinedReal'

    _type_id = 0x0E


class TIRealFraction(TIReal, register=True):
    """
    Parser for real fractions

    A `TIRealFraction` has 8 exponent bits and 14 decimal mantissa digits.
    However, unlike a `TIReal`, the floating point value is automatically converted to an exact fraction on-calc.

    A `TIRealFraction` can be used to form `TIComplexFraction`, `TIComplexPi`, or `TIComplexPiFraction` complex numbers.
    """

    versions = [0x06]

    min_exponent = 0x7C

    is_exact = True

    imag_subtype_id = 0x1B

    _type_id = 0x18

    @Loader[frac.Fraction]
    def load_fraction(self, fraction: frac.Fraction):
        with dec.localcontext() as ctx:
            ctx.prec = 14
            decimal = dec.Decimal(fraction.numerator) / fraction.denominator

        super().load_string(str(decimal))

    def fraction(self) -> frac.Fraction:
        return frac.Fraction(self.decimal()).limit_denominator(10000)

    @Loader[str]
    def load_string(self, string: str):
        self.load_fraction(frac.Fraction(squash(string)))

    def string(self) -> str:
        if self.fraction():
            return "%d / %d" % self.fraction().as_integer_ratio()

        else:
            return "0"


class TIRealRadical(RealEntry, register=True):
    r"""
    Parser for real radicals

    A `TIRealRadical` is an exact sum of two square roots with rational scalars.
    Specifically, a `TIRealRadical` can represent numbers of the form

    $$\frac{\pm a\sqrt{b} \pm c\sqrt{d}}{e}$$

    where all values are non-negative integers. Additionally, $b > d \ge 0$ and $e > 0$.

    Each value is given three nibbles of storage in BCD format.
    Sign information for each radical is stored in an additional nibble.

    A `TIRealRadical` can be used to form `TIComplexRadical` complex numbers.
    """

    flash_only = True

    versions = [0x10]

    min_data_length = 9

    imag_subtype_id = 0x1D

    is_exact = True

    _type_id = 0x1C

    @Section(min_data_length)
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, Bits[0:4])[1:2]
    def sign_type(self) -> int:
        """
        The sign type of the real radical

        If the sign type is odd, the left scalar is negative.
        If the sign type is greater than one, the right scalar is negative.
        """

    @View(calc_data, LeftNibbleBCD)[1:3]
    def denominator(self) -> int:
        """
        The denominator of the real radical
        """

    @View(calc_data, RightNibbleBCD)[3:5]
    def right_scalar(self) -> int:
        """
        The right scalar of the real radical
        """

    @View(calc_data, LeftNibbleBCD)[4:6]
    def left_scalar(self) -> int:
        """
        The left scalar of the real radical
        """

    @View(calc_data, RightNibbleBCD)[6:8]
    def right_radicand(self) -> int:
        """
        The right radicand of the real radical
        """

    @View(calc_data, LeftNibbleBCD)[7:9]
    def left_radicand(self) -> int:
        """
        The left radicand of the real radical
        """

    @property
    def sign(self) -> int:
        return -1 if self.decimal() < 0 else 1

    @Loader[dec.Decimal]
    def load_decimal(self, decimal: dec.Decimal):
        raise NotImplementedError

    def decimal(self) -> dec.Decimal:
        return (self.left_scalar * (-1 if self.sign_type % 2 else 1) * dec.Decimal(self.left_radicand).sqrt() +
                self.right_scalar * (-1 if self.sign_type > 1 else 1) * dec.Decimal(self.right_radicand).sqrt()) \
            / self.denominator

    @Loader[str]
    def load_string(self, string: str):
        if not string:
            self.sign_type, self.denominator = 0, 1
            self.left_scalar, self.left_radicand = 0, 0
            self.right_scalar, self.right_radicand = 0, 0

        # Normalize string
        string = replacer(squash(string), {"(": "", ")": "", "*": "", "~": "-", "sqrt": "√", "-": "+-"}).lstrip("+")

        if "." in string:
            raise ValueError("radical type only accepts integer components")

        if "/" not in string:
            string += "/1"

        if "√" not in string:
            string = "0√1+" + string
            string = string.replace("/", "√1/")

        if "+" not in string:
            string = string.replace("/", "+0√1/")

        # Obtain left component, right component, and denominator
        top, bot = string.split("/")
        left, right = top.split("+")

        if "√" not in left:
            left += "√1"

        if "√" not in right:
            right += "√1"

        # Obtain scalars and radicands
        left_scalar, left_radicand = left.split("√")
        right_scalar, right_radicand = right.split("√")

        left_scalar = int(left_scalar if left_scalar.rstrip("+-") else left_scalar + "1")
        left_radicand = int(left_radicand)
        right_scalar = int(right_scalar if right_scalar.rstrip("+-") else right_scalar + "1")
        right_radicand = int(right_radicand)

        # Normalize radicands
        if left_radicand < 0 or right_radicand < 0:
            raise ValueError("square roots cannot be negative")

        if left_radicand < right_radicand:
            left_scalar, right_scalar = right_scalar, left_scalar
            left_radicand, right_radicand = right_radicand, left_radicand

        elif left_radicand == right_radicand:
            left_scalar += right_scalar
            right_scalar = right_radicand = 0

        denominator = int(bot.lstrip("+"))

        if denominator == 0:
            raise ValueError("denominator must be nonzero")

        # Obtain sign type
        left_scalar, right_scalar = left_scalar * sign(denominator), right_scalar * sign(denominator)
        match sign(left_scalar), sign(right_scalar):
            case 1, -1:
                self.sign_type = 2
            case 1, _:
                self.sign_type = 0
            case -1, 1:
                self.sign_type = 1
            case -1, _:
                self.sign_type = 3

            case _:
                raise ValueError(f"invalid sign type {(sign(left_scalar), sign(right_scalar))}")

        self.denominator = abs(denominator)
        self.left_scalar, self.right_scalar = abs(left_scalar), abs(right_scalar)
        self.left_radicand, self.right_radicand = left_radicand, right_radicand

    def string(self) -> str:
        def reduce(part):
            match [*part]:
                case ["0", "√", *_]:
                    return ""

                case [*_, "√", "0"]:
                    return ""

                case [*_, "√", "1"]:
                    return part[:-2]

                case ["1", "√", *_]:
                    return part[1:]

                case ["-1", "√", *_]:
                    return part[2:]

        left = reduce(f"{self.left_scalar * (-1 if self.sign_type % 2 else 1)}√{self.left_radicand}")
        right = reduce(f"{self.right_scalar * (-1 if self.sign_type > 1 else 1)}√{self.right_radicand}")

        match left, right, self.denominator:
            case "", "", _:
                string = "0"

            case "", _, 1:
                string = right

            case _, "", 1:
                string = left

            case "", _, _:
                string = f"{right} / {self.denominator}"

            case _, "", _:
                string = f"{left} / {self.denominator}"

            case _:
                string = f"({left} + {right}) / {self.denominator}"

        return string.replace("+ -", "- ")


class TIRealPi(TIReal, register=True):
    """
    Parser for real integer multiples of π

    A `TIRealPi` is simply a `TIReal` with an implicit factor of π.

    A `TIRealPi` can be used to form `TIComplexPi` or `TIComplexPiFraction` complex numbers.
    """

    flash_only = True

    versions = [0x10]

    min_exponent = 0x7C

    is_exact = True

    imag_subtype_id = 0x1E

    _type_id = 0x20

    @Loader[dec.Decimal]
    def load_decimal(self, decimal: dec.Decimal):
        raise NotImplementedError

    def decimal(self) -> dec.Decimal:
        with dec.localcontext() as ctx:
            ctx.prec = 14

            return super().decimal() * pi

    @Loader[str]
    def load_string(self, string: str):
        string = replacer(string, {"pi": "π", "*": ""})

        if "π" not in string:
            raise ValueError("value must be a multiple of π")

        super().load_string(string.strip("π"))

    def string(self) -> str:
        string = f"{self.decimal() / pi:.14g}".rstrip("0").rstrip(".")

        if string.startswith("0e"):
            return "0"
        else:
            return string + "π"


class TIRealPiFraction(TIRealPi, TIRealFraction, register=True):
    """
    Parser for real fractional multiples of π

    A `TIRealPiFraction` is simply a `TIRealFraction` with an implicit factor of π.

    A `TIRealPiFraction` can be used to form `TIComplexPiFraction` complex numbers.
    """

    flash_only = True

    imag_subtype_id = 0x1F

    _type_id = 0x21

    def fraction(self) -> frac.Fraction:
        return frac.Fraction(self.decimal() / pi).limit_denominator(10000)

    @Loader[str]
    def load_string(self, string: str):
        replacer(string, {"pi": "π", "*": ""})

        if string != "0" and "π" not in string:
            raise ValueError("value must be a fractional multiple of π")

        super(TIRealPi, self).load_string(string.replace("π", ""))

    def string(self) -> str:
        return super(TIRealPi, self).string().replace(" /", "π /")


__all__ = ["TIReal", "TIUndefinedReal", "TIRealFraction", "TIRealRadical", "TIRealPi", "TIRealPiFraction",
           "RealEntry", "GraphRealEntry"]



import re

from io import BytesIO
from typing import ByteString
from warnings import warn

from tivars.models import *
from tivars.tokenizer import *
from ..data import *
from ..var import SizedEntry


class TokenizedEntry(SizedEntry):
    """
    Base class for all tokenized entries

    A tokenized entry is a `SizedEntry` whose data comprises a stream of tokens.
    """

    versions = [
        0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06,
        0x0A, 0x0B, 0x0C,
        0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26,
        0x2A, 0x2B, 0x2C
    ]

    min_data_length = 2

    clock_tokens = [
        b'\xEF\x00', b'\xEF\x01', b'\xEF\x02', b'\xEF\x03', b'\xEF\x04',
        b'\xEF\x07', b'\xEF\x08', b'\xEF\x09', b'\xEF\x0A', b'\xEF\x0B', b'\xEF\x0C', b'\xEF\x0D',
        b'\xEF\x0E', b'\xEF\x0F', b'\xEF\x10'
    ]
    """
    Tokens which interface with the RTC
    
    These tokens influence the entry's version, though detecting the presence of the RTC has no current application.
    """

    def __format__(self, format_spec: str) -> str:
        if "." not in format_spec:
            format_spec += ".en"

        spec, lang = format_spec.split(".")

        match spec:
            case "":
                return self.decode(self.data, lang=lang)

            case "t":
                return self.decode(self.data, lang=lang, mode="accessible")

            case _:
                return super().__format__(format_spec)

    @staticmethod
    def decode(data: bytes, *, lang: str = "en", mode: str = "display") -> str | bytes:
        """
        Decodes a byte stream into a string of tokens

        Each token is represented using one of three different representations formats, dictated by `mode`:
            - `display`: Represents the tokens with Unicode characters matching the calculator's display
            - `accessible`: Represents the tokens with ASCII-only equivalents, often requiring multi-character glyphs
            - `ti_ascii`: Represents the tokens with their internal font indices (returns a `bytes` object)

        :param data: The bytes to decode
        :param lang: The language used in `string` (defaults to English, `en`)
        :param mode: The form of token representation to use for output (defaults to `display`)
        :return: A string of token representations
        """

        return decode(data, lang=lang, mode=mode)[0]

    @staticmethod
    def encode(string: str, *, model: TIModel = None, lang: str = None) -> bytes:
        """
        Encodes a string of token represented in text into a byte stream

        :param string: The tokens to encode
        :param model: The model to target when encoding (defaults to no specific model)
        :param lang: The language used in `string` (defaults to English, `en`)
        :return: A stream of token bytes
        """

        model = model or TI_84PCE
        return encode(string, model.get_trie(lang))[0]

    def get_min_os(self, data: bytes = None) -> OsVersion:
        return decode(data or self.data)[1]

    def get_version(self, data: bytes = None) -> int:
        match self.get_min_os(data):
            case os if os >= TI_84PCE.OS("5.3"):
                version = 0x0C

            case os if os >= TI_84PCE.OS("5.2"):
                version = 0x0B

            case os if os >= TI_84PCSE.OS("4.0"):
                version = 0x0A

            case os if os >= TI_84P.OS("2.55"):
                version = 0x07

            case os if os >= TI_84P.OS("2.53"):
                version = 0x06

            case os if os >= TI_84P.OS("2.30"):
                version = 0x05

            case os if os >= TI_84P.OS("2.21"):
                version = 0x04

            case os if os >= TI_83P.OS("1.16"):
                version = 0x03

            case os if os >= TI_83P.OS("1.15"):
                version = 0x02

            case os if os >= TI_83P.OS("1.00"):
                version = 0x01

            case _:
                version = 0x00

        if any(token in (data or self.data) for token in self.clock_tokens):
            version += 0x20

        return version

    @Loader[ByteString, BytesIO]
    def load_bytes(self, data: bytes | BytesIO):
        super().load_bytes(data)

        try:
            if self.version != (version := self.get_version()):
                warn(f"The version is incorrect (expected 0x{version:02x}, got 0x{self.version:02x}).",
                     BytesWarning)

        except ValueError as e:
            warn(f"The file contains an invalid token {' '.join(str(e).split()[2:])}.",
                 BytesWarning)

    @Loader[str]
    def load_string(self, string: str, *, model: TIModel = None, lang: str = None):
        self.data = self.encode(string, model=model, lang=lang)

    def string(self) -> str:
        return format(self, "")


class EquationName(TokenizedString):
    """
    Converter for the name section of equations

    Equation names can be any of `Y1` - `Y0`, `X1T` - `X6T`, `Y1T` - `Y6T`, `r1` - `r6`, `u`, `v`, or `w`.
    """

    _T = str

    @classmethod
    def get(cls, data: bytes, **kwargs) -> _T:
        """
        Converts `bytes` -> `str` as done by the memory viewer

        :param data: The raw bytes to convert
        :return: The equation name contained in `data`
        """

        varname = super().get(data)

        if varname.startswith("|"):
            return varname[1:]

        else:
            return varname.upper().strip("{}")

    @classmethod
    def set(cls, value: _T, **kwargs) -> bytes:
        """
        Converts `str` -> `bytes` to match appearance in the memory viewer

        :param value: The value to convert
        :return: The name encoding of `value`
        """

        varname = value[:8].lower()

        if varname.startswith("|") or varname in ("u", "v", "w"):
            varname = "|" + varname[-1]

        elif varname[0] != "{" and varname[-1] != "}":
            varname = "{" + varname + "}"

        return super().set(varname)


class TIEquation(TokenizedEntry, register=True):
    """
    Parser for equations

    A `TIEquation` is a stream of tokens that is evaluated either for graphing or on the homescreen.
    """

    extensions = {
        None: "8xy",
        TI_82: "82y",
        TI_83: "83y",
        TI_83P: "8xy"
    }

    _type_id = 0x03

    def __init__(self, init=None, *,
                 for_flash: bool = True, name: str = "Y1",
                 version: bytes = None, archived: bool = None,
                 data: ByteString = None):

        super().__init__(init, for_flash=for_flash, name=name, version=version, archived=archived, data=data)

    @Section(8, EquationName)
    def name(self) -> str:
        """
        The name of the entry

        Must be one of the equation names: `Y1` - `Y0`, `X1T` - `X6T`, `Y1T` - `Y6T`, `r1` - `r6`, `u`, `v`, or `w`
        """


class TINewEquation(TIEquation, register=True):
    """
    Parser for internal equations

    A `TINewEquation` is simply a `TIEquation` with certain internal uses
    """

    _type_id = 0x0B


class TIString(TokenizedEntry, register=True):
    """
    Parser for strings

    A `TIString` is a stream of tokens.
    """

    extensions = {
        None: "8xs",
        TI_82: "82s",
        TI_83: "83s",
        TI_83P: "8xs"
    }

    _type_id = 0x04

    def __init__(self, init=None, *,
                 for_flash: bool = True, name: str = "Str1",
                 version: bytes = None, archived: bool = None,
                 data: ByteString = None):

        super().__init__(init, for_flash=for_flash, name=name, version=version, archived=archived, data=data)

    @Loader[str]
    def load_string(self, string: str, *, model: TIModel = None):
        super().load_string(string.strip("\"'"))

    def string(self) -> str:
        return f"\"{super().string()}\""


class TIProgram(TokenizedEntry, register=True):
    """
    Parser for programs

    A `TIProgram` is a stream of tokens that is run as a TI-BASIC program.
    """

    extensions = {
        None: "8xp",
        TI_82: "82p",
        TI_83: "83p",
        TI_83P: "8xp"
    }

    is_protected = False
    """
    Whether this program type is protected
    """

    _type_id = 0x05

    @Section(8, TokenizedString)
    def name(self, value) -> str:
        """
        The name of the entry

        Names must 1 to 8 characters in length.
        The name can include any characters A-Z, 0-9, or θ.
        The name cannot start with a digit.
        """

        varname = value[:8].upper()
        varname = re.sub(r"(\u03b8|\u0398|\u03F4|\u1DBF)", "θ", varname)
        varname = re.sub(r"[^θa-zA-Z0-9]", "", varname)

        if not varname or varname[0].isnumeric():
            warn(f"Var has invalid name: {varname}.",
                 BytesWarning)

        return varname


class TIProtectedProgram(TIProgram, register=True):
    """
    Parser for protected programs

    A `TIProtectedProgram` is a `TIProgram` with protection against editing.
    """

    is_protected = True

    _type_id = 0x06


__all__ = ["TIEquation", "TINewEquation", "TIString", "TIProgram", "TIProtectedProgram"]



import decimal as dec

from ..data import *


with dec.localcontext() as ctx:
    ctx.prec = 16
    pi, prev = 3, 0
    t = dec.Decimal(3)

    n, na, d, da = 1, 0, 0, 24
    while pi != prev:
        prev = pi

        n, na = n + na, na + 8
        d, da = d + da, da + 32
        t = (t * n) / d
        pi += t

    ctx.prec = 14


def replacer(string: str, replacements: dict[str, str]) -> str:
    """
    Iteratively applies string replacements

    :param string: The input string
    :param replacements: The replacements to make
    :return: The input string with all replacements made in-order
    """

    for substring, replacement in replacements.items():
        string = string.replace(substring, replacement)

    return string


def sign(x: int):
    """
    Computes the mathematical sign of a number

    :param x: The input number
    :return: The sign of `x`
    """

    return (x > 0) - (x < 0)


def squash(string: str) -> str:
    """
    Removes all spaces from a string

    :param string: The input string
    :return: The input string with all spaces removed
    """

    return ''.join(string.split())


class BCD(Converter):
    """
    Converter for 2-digit binary-coded decimal

    A single byte contains two decimal digits as if they were hex digits.
    """

    _T = int

    @classmethod
    def get(cls, data: bytes, **kwargs) -> _T:
        """
        Converts `bytes` -> `int` from 2-digit binary coded decimal

        :param data: The raw bytes to convert
        :return: The number stored in `data`
        """

        value = 0
        for byte in data:
            value *= 100
            tens, ones = divmod(byte, 16)
            value += 10 * tens + ones

        return value

    @classmethod
    def set(cls, value: _T, **kwargs) -> bytes:
        """
        Converts `int` -> `bytes` as 2-digit binary coded decimal

        :param value: The value to convert
        :return: The bytes representing `value` in BCD
        """

        return int.to_bytes(int(str(value), 16), 7, 'big')


class LeftNibbleBCD(Converter):
    """
    Converter for 2-digit binary-coded decimal with a single extra nibble

    A single byte contains two decimal digits as if they were hex digits.
    The extraneous nibble appears in the leftmost byte, left-padded with a single hex digit.
    """

    _T = int

    @classmethod
    def get(cls, data: bytes, **kwargs) -> _T:
        """
        Converts `bytes` -> `int` from 2-digit binary coded decimal with an extra nibble on the left

        :param data: The raw bytes to convert
        :return: The number stored in `data`
        """

        value = data[0] % 16
        for byte in data[1:]:
            value *= 100
            tens, ones = divmod(byte, 16)
            value += 10 * tens + ones

        return value

    @classmethod
    def set(cls, value: _T, current: bytes = None, **kwargs) -> bytes:
        """
        Converts `int` -> `bytes` as 2-digit binary coded decimal with an extra nibble on the left

        :param value: The value to convert
        :param current: The current value of the data section
        :return: The bytes representing `value` in BCD
        """

        data = bytearray(int.to_bytes(int(str(value), 16), 2, 'big'))
        data[0] += current[0] & 240

        return bytes(data)


class RightNibbleBCD(Converter):
    """
    Converter for 2-digit binary-coded decimal with a single extra nibble

    A single byte contains two decimal digits as if they were hex digits.
    The extraneous nibble appears in the rightmost byte, right-padded with a single hex digit.
    """

    _T = int

    @classmethod
    def get(cls, data: bytes, **kwargs) -> _T:
        """
        Converts `bytes` -> `int` from 2-digit binary coded decimal with an extra nibble on the right

        :param data: The raw bytes to convert
        :return: The number stored in `data`
        """

        value = 0
        for byte in data[:-1]:
            value *= 100
            tens, ones = divmod(byte, 16)
            value += 10 * tens + ones

        return 10 * value + data[-1] // 16

    @classmethod
    def set(cls, value: _T, current: bytes = None, **kwargs) -> bytes:
        """
        Converts `int` -> `bytes` as 2-digit binary coded decimal with an extra nibble on the right

        :param value: The value to convert
        :param current: The current value of the data section
        :return: The bytes representing `value` in BCD
        """

        data = bytearray(int.to_bytes(int(str(10 * value), 16), 2, 'big'))
        data[1] += current[1] % 16

        return bytes(data)


__all__ = ["pi", "replacer", "sign", "squash",
           "BCD", "LeftNibbleBCD", "RightNibbleBCD"]



import functools
import xml.etree.ElementTree as ET
from dataclasses import dataclass

# Models ordered such that models earlier in the list are earlier in the evolution of the token tables
MODEL_ORDER = {
    "": 0,

    "TI-82": 10,

    "TI-83": 20,
    "TI-82ST": 20,
    "TI-82ST.fr": 20,
    "TI-76.fr": 20,

    "TI-83+": 30,
    "TI-83+SE": 30,
    "TI-83+.fr": 30,
    "TI-82+": 30,

    "TI-84+": 40,
    "TI-84+SE": 40,
    "TI-83+.fr:USB": 40,
    "TI-84P.fr": 40,
    "TI-84+PSE": 40,

    "TI-82A": 45,
    "TI-84+T": 45,

    "TI-84+CSE": 50,

    "TI-84+CE": 60,
    "TI-84+CET": 60,
    "TI-83PCE": 60,
    "TI-83PCEEP": 60,
    "TI-84+CEPY": 60,
    "TI-84+CETPE": 60,
    "TI-82AEP": 60,

    "latest": 9999999
}


@functools.total_ordering
@dataclass
class OsVersion:
    """
    Data class for defining and comparing OS versions

    An OS version is defined by its model and verison number.
    The model name must appear in the MODEL_ORDER map above.
    The version number must be of the form "x1.x2.....xn", where each xi is an integer.

    Both the model and version can also be empty or "latest".
    An empty model/version is always first in version ordering, while "latest" is always last.
    """
    
    model: str
    version: str

    def __lt__(self, other):
        order1 = MODEL_ORDER[self.model]
        order2 = MODEL_ORDER[other.model]

        if order1 < order2:
            return True
        elif order1 > order2:
            return False
        else:
            if self.version == "latest":
                return False
            elif other.version == "latest":
                return True
            elif other.version == "":
                return False
            elif self.version == "":
                return True
            else:
                return any(map(lambda a: a[0] < a[1],
                               zip(map(int, self.version.split(".")), map(int, other.version.split(".")))))

    def __eq__(self, other):
        return MODEL_ORDER[self.model] == MODEL_ORDER[other.model] and self.version == other.version

    @staticmethod
    def from_element(element) -> 'OsVersion':
        """
        Constructs an instance from an XML element in a token sheet

        :param element: An XML element
        :return: An OS version corresponding to the element
        """
        
        model = ""
        version = ""

        for child in element:
            if child.tag == "model":
                model = child.text
            elif child.tag == "os-version":
                version = child.text
            else:
                raise ValueError("Unrecognized tag in " + element.tag + ": " + child.tag)

        if version == "":
            raise ValueError("<" + element.tag + "> has a missing or empty <os-version> tag.")
        elif model == "":
            raise ValueError("<" + element.tag + "> has a missing or empty <model> tag.")

        if model not in MODEL_ORDER or model == "latest":  # "latest" is for user convenience, not the sheet itself
            raise ValueError("Unrecognized <model>: " + model)

        if any([c != '.' and not c.isnumeric() for c in version]):
            raise ValueError(
                "Invalid <version> string \"" + version + "\", must be a sequence of numbers separated by periods.")

        return OsVersion(model, version)


class OsVersions:
    """
    Enum class to contain useful OS version constants

    This class can be extended with useful versions for other applications.
    """
    
    INITIAL = OsVersion("", "")
    LATEST = OsVersion("latest", "latest")


class Translation:
    """
    Data class for managing the different decodings of tokens

    A token translations stores the following:
        - The TI ASCII representation, i.e. the sequence of font bytes used on-calc
        - The displayed name, a Unicode string intended to best emulate on-calc display
        - The accessible name, an ASCII string intended to be easy to type
        - Any variant names; such may be derived from their use in other tokenization tools
    """
    
    def __init__(self, ti_ascii: bytes, display: str, accessible: str, variants: list[str]):
        self.ti_ascii = ti_ascii
        self.display = display
        self.accessible = accessible
        self.variants = variants

    def names(self) -> list[str]:
        """
        :return: A list of all names for this token
        """
        
        return [self.display, self.accessible] + self.variants

    @staticmethod
    def from_element(element) -> (str, 'Translation'):
        """
        Constructs an instance and its key from an XML element in a token sheet

        :param element: An XML element
        :return: A tuple of a string key and a token translation corresponding to the element
        """
        
        code = element.attrib["code"]

        ti_ascii = bytes.fromhex(element.attrib["ti-ascii"])

        display = ""
        accessible = ""
        variants = []

        for child in element:
            match child.tag:
                case "display":
                    display = child.text
                case "accessible":
                    accessible = child.text
                case "variants":
                    variants.append(child.text)

        return code, Translation(ti_ascii, display, accessible, variants)


class Token:
    """
    Data class for storing all information about a single token

    A token stores the following:
        - The bytes for this token on-calc
        - The translations of this tokens in all supported languages
        - The earliest OS version supporting this token
        - The latest OS version supporting this token
        - Any additional attributes stored in the token sheets
    """
    
    def __init__(self, bits: bytes, langs: dict[str, Translation], attrs: dict[str, str] = None,
                 since: OsVersion = OsVersions.INITIAL,
                 until: OsVersion = OsVersions.LATEST):
        self.bits = bits
        self.langs = langs
        self.attrs = attrs
        self.since = since
        self.until = until

    @staticmethod
    def from_element(element, bits, version=OsVersions.LATEST):
        """
        Constructs an instance from an XML element in the token sheets

        :param element: An XML element
        :param bits: The token's bytes
        :param version: A minimum OS version (defaults to latest)
        :return: A token corresponding to the element and bits
        """
        
        since = OsVersions.INITIAL
        until = OsVersions.LATEST

        langs: dict[str, Translation] = {}

        done = False
        for version_elem in element:
            for child in version_elem:
                match child.tag:
                    case "since":
                        version_since = OsVersion.from_element(child)
                        if since < version_since:
                            since = version_since

                        if since > version:
                            done = True

                    case "until":
                        version_until = OsVersion.from_element(child)
                        if until > version_until:
                            until = version_until

                    case "lang":
                        if not done:
                            code, translation = Translation.from_element(child)
                            langs[code] = translation
        return Token(bits, langs, attrs=element.attrib, since=since, until=until)


class Tokens:
    """
    Data class for storing maps between text and tokens

    Two maps are stored: bytes -> token and string -> string -> bytes.
    
    The byte map is indexed by bytes and yields token objects.
    The lang map is indexed by language code, then token name, and yields token bytes.
    """
    
    def __init__(self, byte_map: dict[bytes, Token], lang_map: dict[str, dict[str, bytes]]):
        self.bytes = byte_map
        self.langs = lang_map

    @staticmethod
    def from_xml_string(xml_str: str, version=OsVersions.LATEST):
        """
        Constructs an instance from an XML string

        :param xml_str: An XML string
        :param version: A minimum OS version (defaults to latest)
        :return: Token maps corresponding to the string
        """
        
        return Tokens.from_element(ET.fromstring(xml_str), version=version)

    @staticmethod
    def from_element(root, version=OsVersions.LATEST):
        """
        Constructs an instance from an XML element in the token sheets

        :param root: An XML element, which must be the root element of the sheet
        :param version: A minimum OS version (defauls to latest)
        :return: Token maps corresponding to the root element
        """
        
        if root.tag != "tokens":
            raise ValueError("Not a tokens xml.")

        all_bytes: dict[bytes, Token] = {}
        all_langs: dict[str, dict[str, bytes]] = {}

        def parse_page(element, bits=b""):
            nonlocal all_bytes
            nonlocal all_langs

            if element.tag == "token":
                token_bits = bits + bytes.fromhex(element.attrib["value"][1:])
                token = Token.from_element(element, token_bits, version=version)

                if token.langs:
                    all_bytes[token_bits] = token
                    for lang, translation in token.langs.items():
                        if lang not in all_langs:
                            all_langs[lang] = {}

                        for name in translation.names():
                            all_langs[lang][name] = token_bits

            for child in element:
                if child.tag == "two-byte":
                    parse_page(child, bits=bits + bytes.fromhex(child.attrib["value"][1:]))
                else:
                    parse_page(child, bits=bits)

        parse_page(root)

        return Tokens(all_bytes, all_langs)


# with open("../8X.xml", encoding="UTF-8") as file:
#   Tokens.from_xml_string(file.read())



from .parse import Token, Tokens


class TokenTrie:
    """
    Basic trie class for tokenizing text
    """
    
    def __init__(self, token: Token = None):
        self.token = token
        self.children = {}

    def insert(self, token: Token, lang: str = "en", *, mode: str = "all"):
        """
        Inserts some or all names of a token into the trie in a given language

        Three modes can be passed to specify which name(s) of the tokens to insert:
            - "display":      insert only the display name
            - "accessible":   insert only the accessible name
            - "all":          insert all names (display, accessible, and variants)

        :param token: The token to insert
        :param lang: The language to insert names from
        :param mode: Type(s) of names to insert (defaults to "all")
        """
        
        if lang not in token.langs:
            raise ValueError(f"lang {lang} not found")

        match mode.lower():
            case "display":
                names = [token.langs[lang].display]

            case "accessible":
                names = [token.langs[lang].accessible]

            case _:
                # Case "all", but can be any unrecognized arg
                names = token.langs[lang].names()

        for name in names:
            current = self
            for char in name:
                if char not in current.children:
                    current.children[char] = TokenTrie()

                current = current.children[char]

            current.token = token

    @staticmethod
    def from_tokens(tokens: Tokens, lang: str = "en", *, mode: str = "all"):
        """
        Inserts all tokens from a token container into the trie

        The lang and mode parameters match those used by a single insert.
        """
        
        if lang not in tokens.langs:
            raise ValueError(f"lang {lang} not found")

        root = TokenTrie()
        for token in tokens.bytes.values():
            root.insert(token, lang, mode=mode)

        return root

    def get_tokens(self, string: str) -> list[tuple[Token, str]]:
        """
        Finds all tokens which can be parsed from a given input string

        Each token is returned with the portion of the input string still remaining.
        Output is sorted by decreasing length of the consumed input.

        :return: A list of tuples each containing a token and its remaining input
        """
        
        tokens = []

        if string and string[0] in self.children:
            tokens += self.children[string[0]].get_tokens(string[1:])

        if self.token:
            tokens.append((self.token, string))

        return tokens

    def get_longest_token(self, string: str) -> tuple[Token, str]:
        """
        Finds the longest token which can be parsed from a given input string

        :return: A tuple of a token and the remaining input after parsing
        """
        
        return self.get_tokens(string)[0]



import io
import json
import os

from typing import ByteString, Iterator
from warnings import warn

from tivars.models import *
from ..flags import *
from ..data import *
from ..var import TIEntry, SizedEntry
from .real import *
from .tokenized import TIEquation


class GraphMode(Flags):
    Dot = {0: 1}
    Connected = {0: 0}
    Simul = {1: 1}
    Sequential = {1: 0}
    GridOn = {2: 1, 7: 0}
    GridOff = {2: 0, 7: 0}
    GridLine = {2: 1, 7: 1}
    GridDot = {2: 1, 7: 0}
    PolarGC = {3: 1}
    RectGC = {3: 0}
    CoordOn = {4: 0}
    CoordOff = {4: 1}
    AxesOff = {5: 1}
    AxesOn = {5: 0}
    LabelOn = {6: 1}
    LabelOff = {6: 0}

    ExprOff = {0: 1}
    ExprOn = {0: 0}
    SEQ_n = {1: 0, 2: 0}
    SEQ_np1 = {1: 1, 2: 0}
    SEQ_np2 = {1: 0, 2: 1}

    DetectAsymptotesOff = {0: 1}
    DetectAsymptotesOn = {0: 0}


class SeqMode(Flags):
    Time = {0: 0, 1: 0, 2: 0, 3: 0, 4: 0}
    Web = {0: 1, 1: 0, 2: 0, 3: 0, 4: 0}
    VertWeb = {0: 0, 1: 1, 2: 0, 3: 0, 4: 0}
    uv = {0: 0, 1: 0, 2: 1, 3: 0, 4: 0}
    vw = {0: 0, 1: 0, 2: 0, 3: 1, 4: 0}
    uw = {0: 0, 1: 0, 2: 0, 3: 0, 4: 1}


class GraphStyle(Enum):
    SolidLine = b'\x00'
    ThickLine = b'\x01'
    ShadeAbove = b'\x02'
    ShadeBelow = b'\x03'
    Trace = b'\x04'
    Animate = b'\x05'
    DottedLine = b'\x06'

    _all = [SolidLine, ThickLine, ShadeAbove, ShadeBelow, Trace, Animate, DottedLine]
    STYLES = _all[:7]


class GraphColor(Enum):
    Mono = b'\x00'
    Blue = b'\x01'
    Red = b'\x02'
    Black = b'\x03'
    Magenta = b'\x04'
    Green = b'\x05'
    Orange = b'\x06'
    Brown = b'\x07'
    Navy = b'\x08'
    LtBlue = b'\x09'
    Yellow = b'\x0A'
    White = b'\x0B'
    LtGray = b'\x0C'
    MedGray = b'\x0D'
    Gray = b'\x0E'
    DarkGray = b'\x0F'

    _all = [Mono, Blue, Red, Black, Magenta, Green, Orange, Brown, Navy,
            LtBlue, Yellow, White, LtGray, MedGray, Gray, DarkGray]
    COLORS = _all[1:]


class GlobalLineStyle(Enum):
    Thick = b'\x00'
    DotThick = b'\x01'
    Thin = b'\x02'
    DotThin = b'\x03'

    _all = [Thick, Thin, DotThick, DotThin]
    STYLES = _all


class BorderColor(Enum):
    LtGray = b'\x01'
    Teal = b'\x02'
    LtBlue = b'\x03'
    White = b'\x04'

    _all = [LtGray, Teal, LtBlue, White]
    COLORS = _all


class EquationFlags(Flags):
    Selected = {5: 1}
    Deselected = {5: 0}
    UsedForGraph = {6: 1}
    UnusedForGraph = {6: 0}
    LinkTransferSet = {7: 1}
    LinkTransferClear = {7: 0}


class TIGraphedEquation(TIEquation):
    min_data_length = 3

    class Raw(TIEntry.Raw):
        __slots__ = "meta_length", "type_id", "name", "version", "archived", "style", "color", "calc_data"

    def __init__(self, init=None, *,
                 for_flash: bool = True, name: str = "Y1",
                 version: bytes = None, archived: bool = None,
                 data: ByteString = None):
        super().__init__(init, for_flash=for_flash, name=name, version=version, archived=archived, data=data)

        self.flags = EquationFlags({0: 1, 1: 1})
        self.style = b'\x00'
        self.color = b'\x00'

    def __class_getitem__(cls, index: int):
        index -= 1

        class IndexedEquationConverter(Converter):
            """
            Converter for equations within a GDB

            Since equations are stored contiguously within a GDB, their exact location is not static.
            This converter interfaces with the nth equation by counting up from the start of the data section.

            The `n`th equation in a GDB is interfaced by the converter `TIGraphedEquation[n]`.
            """

            _T = TIGraphedEquation

            @classmethod
            def get(cls, data: bytes, *, instance: 'TIMonoGDB' = None, **kwargs) -> _T:
                """
                Converts `bytes` -> `TIGraphedEquation` by finding the equation at `index` within a GDB

                :param data: The raw bytes to convert
                :param instance: The instance which contains the data section
                :return: The bytes in `data`, unchanged
                """

                return instance.equations[index]

            @classmethod
            def set(cls, value: _T, *, instance: 'TIMonoGDB' = None, **kwargs) -> bytes:
                """
                Converts `bytes` -> `TIGraphedEquation` by modifying the equation at `index` within a GDB

                :param value: The value to convert
                :param instance: The instance which contains the data section
                :return: The bytes in `value`, unchanged
                """

                # Set appropriate equation
                equations = list(instance.equations)
                equations[index] = value

                # Set styles
                data = instance.raw.calc_data[:instance.offset]
                for i in range(0, instance.num_equations, instance.num_equations // instance.num_styles):
                    data += equations[i].style

                # Set data
                data += b''.join(equation.raw.calc_data for equation in equations)

                # Set colors (if they exist)
                if color := color_data(instance):
                    data += b'84C'
                    for i in range(0, instance.num_equations, instance.num_equations // instance.num_styles):
                        data += equations[i].color

                    data += color[-5:]

                return data

        return IndexedEquationConverter

    def __iter__(self) -> Iterator:
        return iter(self.dict().items())

    @Section(1, GraphStyle)
    def style(self) -> bytes:
        """
        The style of the GDB equation

        This value is not intrinsically stored with the equation, but is bundled for convenience.
        """

    @Section(1, GraphColor)
    def color(self) -> bytes:
        """
        The color of the GDB equation

        This value is not intrinsically stored with the equation, but is bundled for convenience.
        """

    @Section()
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, EquationFlags)[0:1]
    def flags(self) -> EquationFlags:
        """
        The flags for the GDB equation

        The flags track whether the equation is selected, used for graphing, or is participating in a link transfer.
        """

    @View(calc_data, Integer)[1:3]
    def length(self) -> int:
        """
        The length of this entry's user data section
        """

    def load_data_section(self, data: io.BytesIO):
        flag_byte = data.read(1)
        data_length = int.from_bytes(length_bytes := data.read(2), 'little')
        self.raw.calc_data = bytearray(flag_byte + length_bytes + data.read(data_length))

    @Loader[dict]
    def load_dict(self, dct: dict):
        """
        Loads a JSON `dict` into this GDB equation

        :param dct: The dict to load
        """

        self.style = getattr(GraphStyle, dct["style"])
        self.color = getattr(GraphColor, dct["color"])

        self.load_equation(TIEquation(dct["expr"]))

        # Set flags
        flags = {"selected": False, "wasUsedForGraph": False, "linkTransfer": False} | dct["flags"]
        self.flags = EquationFlags({1: 1, 0: 1})

        self.flags |= EquationFlags.Selected if flags["selected"] else EquationFlags.Deselected
        self.flags |= EquationFlags.UsedForGraph if flags["wasUsedForGraph"] else EquationFlags.UnusedForGraph
        self.flags |= EquationFlags.LinkTransferSet if flags["linkTransfer"] else EquationFlags.LinkTransferClear

    def dict(self) -> dict:
        """
        :return: A `dict` representing this GDB equation in JSON format
        """

        dct = {"style": GraphStyle.get_name(self.style)}

        if self.color != GraphColor.Mono:
            dct["color"] = GraphColor.get_name(self.color)

        return dct | {
            "flags": {
                "selected": EquationFlags.Selected in self.flags,
                "wasUsedForGraph": EquationFlags.UsedForGraph in self.flags,
                "linkTransfer": EquationFlags.LinkTransferSet in self.flags
            },
            "expr": self.string()
        }

    @Loader[TIEquation]
    def load_equation(self, equation: TIEquation):
        """
        Loads a `TIEquation` into this GDB equation

        :param equation: The equation to load
        """

        self.raw.calc_data[1:] = equation.calc_data

    def equation(self) -> TIEquation:
        """
        :return: The `TIEquation` component of this GDB equation
        """

        return TIEquation(self.bytes()[:-self.data_length] + self.bytes()[-self.data_length + 1:])

    @Loader[str]
    def load_string(self, string: str, *, model: TIModel = None):
        equation = TIEquation()
        equation.load_string(string, model=model)
        self.load_equation(equation)


def color_data(gdb: 'TIMonoGDB') -> bytes:
    data = io.BytesIO(gdb.calc_data[gdb.offset + gdb.num_styles:])
    for i in range(gdb.num_equations):
        TIGraphedEquation().load_data_section(data)

    return data.read()


class TIMonoGDB(SizedEntry, register=True):
    """
    Base class for all GDB entries

    A GDB is a collection of equations and graph settings representing the state of one of the equation plotters.
    A GDB can correspond to one of the Function, Parametric, Polar, or Sequence plotting modes.
    GDBs for monochrome models are truncations of those for color models.
    """

    extensions = {
        None: "8xd",
        TI_82: "82d",
        TI_83: "83d",
        TI_83P: "8xd"
    }

    mode_byte = 0x00
    """
    The byte which identifies the GDB type
    """

    min_data_length = 61
    has_color = False
    """
    Whether this GDB type carries color information
    """

    num_equations = 0
    """
    The number of equations contained in this GDB type
    """

    num_parameters = 0
    """
    The number of graph parameters contained in this GDB type
    """

    num_styles = 0
    """
    The number of equation styles contained in this GDB type
    """

    equation_names = []
    """
    The names of the equations in this GDB type
    """

    _type_id = 0x08

    def __init__(self, init=None, *,
                 for_flash: bool = True, name: str = "GDB1",
                 version: bytes = None, archived: bool = None,
                 data: ByteString = None):

        super().__init__(init, for_flash=for_flash, name=name, version=version, archived=archived, data=data)

    def __iter__(self) -> Iterator:
        return iter(self.dict().items())

    @Section()
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, Integer)[3:4]
    def mode_id(self) -> int:
        """
        The mode ID for the GDB

        The ID is one of `0x10` (Function), `0x20` (Polar), `0x40` (Parametric), or `0x80` (Sequence).
        """

    @View(calc_data, GraphMode)[4:5]
    def mode_flags(self) -> GraphMode:
        """
        The flags for the mode settings

        Dot/Connected, Simul/Sequential, GridOn/Line/Dot/Off, PolarGC/RectGC, CoordOn/Off, AxesOff/On, and LabelOn/Off
        """

    @View(calc_data, GraphMode)[6:7]
    def extended_mode_flags(self) -> GraphMode:
        """
        The flags for the extended mode settings

        ExprOn/Off and sequence plot offsets for sequence mode
        """

    @View(calc_data, GraphRealEntry)[7:16]
    def Xmin(self) -> GraphRealEntry:
        """
        Xmin: the X-coordinate of the left edge of the graphscreen
        """

    @View(calc_data, GraphRealEntry)[16:25]
    def Xmax(self) -> GraphRealEntry:
        """
        Xmax: the X-coordinate of the right edge of the graphscreen
        """

    @View(calc_data, GraphRealEntry)[25:34]
    def Xscl(self) -> GraphRealEntry:
        """
        Xscl: the separation between ticks on the X-axis
        """

    @View(calc_data, GraphRealEntry)[34:43]
    def Ymin(self) -> GraphRealEntry:
        """
        Ymin: the Y-coordinate of the bottom edge of the graphscreen
        """

    @View(calc_data, GraphRealEntry)[43:52]
    def Ymax(self) -> GraphRealEntry:
        """
        Ymax: the Y-coordinate of the top edge of the graphscreen
        """

    @View(calc_data, GraphRealEntry)[52:61]
    def Yscl(self) -> GraphRealEntry:
        """
        Yscl: the separation between ticks on the Y-axis
        """

    @property
    def mode(self) -> str:
        """
        The mode for the GDB

        One of Function, Parametric, Polar, or Sequence
        """

        match self.mode_id:
            case TIMonoFuncGDB.mode_byte:
                return 'Function'
            case TIMonoParamGDB.mode_byte:
                return 'Parametric'
            case TIMonoPolarGDB.mode_byte:
                return 'Polar'
            case TIMonoSeqGDB.mode_byte:
                return 'Sequence'

            case _:
                warn(f"Graphing mode byte 0x{self.mode_id:x} not recognized.",
                     BytesWarning)

    @property
    def offset(self) -> int:
        """
        :return: The index of the start of the equation styles in this GDB
        """

        return TIMonoGDB.min_data_length + GraphRealEntry.min_data_length * self.num_parameters

    @property
    def equations(self) -> tuple[TIGraphedEquation, ...]:
        """
        :return: This GDB's stored equations
        """

        data = io.BytesIO(self.calc_data[self.offset:])
        equations = tuple(TIGraphedEquation(name=name) for name in self.equation_names)

        # Load styles
        for i in range(self.num_styles):
            style = data.read(1)
            for j in range(r := self.num_equations // self.num_styles):
                equations[r * i + j].style = style

        # Load data sections
        for i in range(self.num_equations):
            equations[i].load_data_section(data)

        # Load colors (if they exist)
        if rest := data.read():
            data = io.BytesIO(rest)
            data.seek(3, 1)

            for i in range(self.num_styles):
                color = data.read(1)
                for j in range(r := self.num_equations // self.num_styles):
                    equations[r * i + j].color = color

        return equations

    def get_min_os(self, data: bytes = None) -> OsVersion:
        return max(map(TIGraphedEquation.get_min_os, self.equations), default=OsVersions.INITIAL)

    def get_version(self, data: bytes = None) -> int:
        return max(map(TIGraphedEquation.get_version, self.equations), default=0x00)

    @Loader[dict]
    def load_dict(self, dct: dict):
        """
        Loads a JSON `dict` into this GDB

        :param dct: The dict to load
        """

        self.clear()
        self.raw.calc_data[3] = {
            'Function': 0x10,
            'Parametric': 0x40,
            'Polar': 0x20,
            'Sequence': 0x80
        }.get(mode := dct.get("graphMode", "Function"), 0x00)

        # Load formatSettings
        for setting in dct.get("formatSettings", []):
            try:
                self.mode_flags |= getattr(GraphMode, setting)
            except AttributeError:
                warn(f"Unrecognized format setting ({setting}).",
                     UserWarning)

        # Load extSettings
        ext_settings = dct.get("extSettings", {})
        if "showExpr" in ext_settings:
            self.extended_mode_flags |= GraphMode.ExprOn if ext_settings["showExpr"] else GraphMode.ExprOff

        if self.raw.calc_data[3] != 0x80:
            if "seqMode" in ext_settings or "seqSettings" in dct:
                warn(f"Sequence settings have been provided, but this GDB is for {mode.lower()} graphs.",
                     UserWarning)

        # Load globalWindowSettings
        for var, value in dct.get("globalWindowSettings", {}).items():
            if not hasattr(self, var):
                warn(f"Unrecognized window setting ({var}).",
                     UserWarning)
            else:
                setattr(self, var, TIReal(value))

        # Load specific data
        data = dct.get("specificData", {})
        for var, value in data.get("settings", {}).items():
            if not hasattr(self, var):
                warn(f"Unrecognized window setting ({var}).",
                     UserWarning)
            else:
                setattr(self, var, TIReal(value))

        # Load equations
        for name, equation in data.get("equations", {}).items():
            if name in self.equation_names:
                plotted = TIGraphedEquation(name=name)
                plotted.load_dict(equation)
                setattr(self, name, plotted)

            else:
                warn(f"Unrecognized equation ({name}).",
                     UserWarning)

        # Set type if color data exists so it can be loaded
        if "global84CSettings" in dct:
            self.__class__ = TIGDB

        self.coerce()
        self._load_dict(dct)

    def _load_dict(self, dct: dict):
        pass

    def dict(self) -> dict:
        """
        :return: A `dict` representing this GDB in JSON format
        """

        return {
            "graphMode": self.mode,
            "formatSettings": [
                "Dot" if GraphMode.Dot in self.mode_flags else "Connected",
                "Simul" if GraphMode.Simul in self.mode_flags else "Sequential",
                "GridOn" if GraphMode.GridOn in self.mode_flags else "GridOff",
                "PolarGC" if GraphMode.PolarGC in self.mode_flags else "RectGC",
                "CoordOff" if GraphMode.CoordOff in self.mode_flags else "CoordOn",
                "AxesOff" if GraphMode.AxesOff in self.mode_flags else "AxesOn",
                "LabelOn" if GraphMode.LabelOn in self.mode_flags else "LabelOff"
            ],
            "extSettings": {
                "showExpr": GraphMode.ExprOn in self.extended_mode_flags
            },
            "globalWindowSettings": {
                "Xmin": self.Xmin.json_number(),
                "Xmax": self.Xmax.json_number(),
                "Xscl": self.Xscl.json_number(),
                "Ymin": self.Ymin.json_number(),
                "Ymax": self.Ymax.json_number(),
                "Yscl": self.Yscl.json_number()
            }
        }

    @Loader[str]
    def load_string(self, string: str):
        self.load_dict(json.loads(string))

    def string(self) -> str:
        return json.dumps(self.dict())

    def coerce(self):
        if color_data(self):
            self.__class__ = TIGDB
            self.coerce()
        else:
            match self.mode_id:
                case TIMonoFuncGDB.mode_byte:
                    self.__class__ = TIMonoFuncGDB
                case TIMonoParamGDB.mode_byte:
                    self.__class__ = TIMonoParamGDB
                case TIMonoPolarGDB.mode_byte:
                    self.__class__ = TIMonoPolarGDB
                case TIMonoSeqGDB.mode_byte:
                    self.__class__ = TIMonoSeqGDB

                case _:
                    warn(f"Graphing mode byte 0x{self.mode_id:x} not recognized.",
                         BytesWarning)


class TIGDB(TIMonoGDB):
    min_data_length = 66
    has_color = True

    def __init__(self, init=None, *,
                 for_flash: bool = True, name: str = "GDB1",
                 version: bytes = None, archived: bool = None,
                 data: ByteString = None):

        super().__init__(init, for_flash=for_flash, name=name, version=version, archived=archived, data=data)

        self.axes_color = GraphColor.Black
        self.grid_color = GraphColor.MedGray

    @Section()
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, GraphColor)[-5:-4]
    def grid_color(self) -> bytes:
        """
        The color of the grid
        """

    @View(calc_data, GraphColor)[-4:-3]
    def axes_color(self) -> bytes:
        """
        The color of the axes
        """

    @View(calc_data, GlobalLineStyle)[-3:-2]
    def global_line_style(self) -> bytes:
        """
        The line style for all plotted equations
        """

    @View(calc_data, BorderColor)[-2:-1]
    def border_color(self) -> bytes:
        """
        The color of the graph border
        """

    @View(calc_data, GraphMode)[-1:]
    def color_mode_flags(self) -> GraphMode:
        """
        The flags for extended color mode settings

        Only DetectAsymptotesOn/Off is stored here.
        """

    def _load_dict(self, dct: dict):
        self.color_magic = "84C"

        if colors := dct.get("global84CSettings", {}).get("colors", {}):
            if "grid" in colors:
                self.grid_color = getattr(GraphColor, colors["grid"])

            if "axes" in colors:
                self.axes_color = getattr(GraphColor, colors["axes"])

            if "border" in colors:
                self.border_color = BorderColor.COLORS[colors["border"] - 1]

        if other := dct.get("global84CSettings", {}).get("other", {}):
            if "globalLineStyle" in other:
                self.global_line_style = getattr(GlobalLineStyle, other["globalLineStyle"])

            if "detectAsymptotes" in other:
                self.color_mode_flags |= \
                    GraphMode.DetectAsymptotesOn if other["detectAsymptotes"] else GraphMode.DetectAsymptotesOff

    def dict(self) -> dict:
        return {
            "global84CSettings": {
                "colors": {
                    "grid": GraphColor.get_name(self.grid_color),
                    "axes": GraphColor.get_name(self.axes_color),
                    "border": self.border_color[0]
                },
                "other": {
                    "globalLineStyle": GlobalLineStyle.get_name(self.global_line_style),
                    "detectAsymptotes": GraphMode.DetectAsymptotesOn in self.color_mode_flags
                }
            }
        }

    def coerce(self):
        match self.mode_id:
            case TIFuncGDB.mode_byte:
                self.__class__ = TIFuncGDB
            case TIParamGDB.mode_byte:
                self.__class__ = TIParamGDB
            case TIPolarGDB.mode_byte:
                self.__class__ = TIPolarGDB
            case TISeqGDB.mode_byte:
                self.__class__ = TISeqGDB

            case _:
                warn(f"Graphing mode byte 0x{self.mode_id:x} not recognized.",
                     BytesWarning)


class TIMonoFuncGDB(TIMonoGDB):
    mode_byte = 0x10

    min_data_length = 110

    num_equations = 10
    num_parameters = 1
    num_styles = 10

    equation_names = ["Y1", "Y2", "Y3", "Y4", "Y5", "Y6", "Y7", "Y8", "Y9", "Y0"]

    @Section()
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, Integer)[3:4]
    def mode_id(self) -> int:
        """
        The mode ID for the GDB - `0x10`
        """

    @View(calc_data, GraphRealEntry)[61:70]
    def Xres(self, value: GraphRealEntry) -> GraphRealEntry:
        """
        Xres: The pixel separation of points in a function plot

        The value must be an integer between 1 and 8.
        """

        if int(value) != float(value) or not 1 <= int(value) <= 8:
            warn(f"Expected an integer between 1 and 8, got {float(value)}.",
                 UserWarning)

        return value

    @View(calc_data, TIGraphedEquation[1])
    def Y1(self) -> TIGraphedEquation:
        """
        Y1: The 1st equation in function mode
        """

    @View(calc_data, TIGraphedEquation[2])
    def Y2(self) -> TIGraphedEquation:
        """
        Y2: The 2nd equation in function mode
        """

    @View(calc_data, TIGraphedEquation[3])
    def Y3(self) -> TIGraphedEquation:
        """
        Y3: The 3rd equation in function mode
        """

    @View(calc_data, TIGraphedEquation[4])
    def Y4(self) -> TIGraphedEquation:
        """
        Y4: The 4th equation in function mode
        """

    @View(calc_data, TIGraphedEquation[5])
    def Y5(self) -> TIGraphedEquation:
        """
        Y5: The 5th equation in function mode
        """

    @View(calc_data, TIGraphedEquation[6])
    def Y6(self) -> TIGraphedEquation:
        """
        Y6: The 6th equation in function mode
        """

    @View(calc_data, TIGraphedEquation[7])
    def Y7(self) -> TIGraphedEquation:
        """
        Y7: The 7th equation in function mode
        """

    @View(calc_data, TIGraphedEquation[8])
    def Y8(self) -> TIGraphedEquation:
        """
        Y8: The 8th equation in function mode
        """

    @View(calc_data, TIGraphedEquation[9])
    def Y9(self) -> TIGraphedEquation:
        """
        Y9: The 9th equation in function mode
        """

    @View(calc_data, TIGraphedEquation[10])
    def Y0(self) -> TIGraphedEquation:
        """
        Y0: The 10th equation in function mode
        """

    @Loader[dict]
    def load_dict(self, dct: dict = None):
        if dct is None:
            with open(os.path.join(os.path.dirname(__file__), "json/func.default.json")) as file:
                dct = json.load(file)

        super().load_dict(dct)

    def dict(self) -> dict:
        return super().dict() | {
            "specificData": {
                "settings": {
                    "Xres": int(self.Xres)
                },
                "equations": {
                    equation.name: equation.dict() for equation in self.equations
                }
            }
        }


class TIFuncGDB(TIGDB, TIMonoFuncGDB):
    min_data_length = 128

    @Section()
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, String)[-18:-15]
    def color_magic(self) -> str:
        """
        Magic to identify the GDB as color-oriented - `84C`
        """

    def mono(self) -> TIMonoFuncGDB:
        """
        :return: A copy of this GDB with all color data removed
        """

        return TIMonoFuncGDB(self.bytes()[:-18])

    def _load_dict(self, dct: dict):
        super(TIGDB, self)._load_dict(dct)
        super(TIFuncGDB, self)._load_dict(dct)

    def dict(self) -> dict:
        return super(TIGDB, self).dict() | super(TIFuncGDB, self).dict()


class TIMonoParamGDB(TIMonoGDB):
    mode_byte = 0x40

    min_data_length = 130

    num_equations = 12
    num_parameters = 3
    num_styles = 6

    equation_names = ["X1T", "Y1T", "X2T", "Y2T", "X3T", "Y3T", "X4T", "Y4T", "X5T", "Y5T", "X6T", "Y6T"]

    @Section()
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, Integer)[3:4]
    def mode_id(self) -> int:
        """
        The mode ID for the GDB - `0x40`
        """

    @View(calc_data, GraphRealEntry)[61:70]
    def Tmin(self) -> GraphRealEntry:
        """
        Tmin: The initial time
        """

    @View(calc_data, GraphRealEntry)[70:79]
    def Tmax(self) -> GraphRealEntry:
        """
        Tmax: The final time
        """

    @View(calc_data, GraphRealEntry)[79:88]
    def Tstep(self) -> GraphRealEntry:
        """
        Tstep: The time increment
        """

    @View(calc_data, TIGraphedEquation[1])
    def X1T(self) -> TIGraphedEquation:
        """
        X1T: The 1st X-component in parametric mode
        """

    @View(calc_data, TIGraphedEquation[2])
    def Y1T(self) -> TIGraphedEquation:
        """
        Y1T: The 1st Y-component in parametric mode
        """

    @View(calc_data, TIGraphedEquation[3])
    def X2T(self) -> TIGraphedEquation:
        """
        X2T: The 2nd X-component in parametric mode
        """

    @View(calc_data, TIGraphedEquation[4])
    def Y2T(self) -> TIGraphedEquation:
        """
        Y2T: The 2nd Y-component in parametric mode
        """

    @View(calc_data, TIGraphedEquation[5])
    def X3T(self) -> TIGraphedEquation:
        """
        X3T: The 3rd X-component in parametric mode
        """

    @View(calc_data, TIGraphedEquation[6])
    def Y3T(self) -> TIGraphedEquation:
        """
        Y3T: The 3rd Y-component in parametric mode
        """

    @View(calc_data, TIGraphedEquation[7])
    def X4T(self) -> TIGraphedEquation:
        """
        X4T: The 4th X-component in parametric mode
        """

    @View(calc_data, TIGraphedEquation[8])
    def Y4T(self) -> TIGraphedEquation:
        """
        Y4T: The 4th Y-component in parametric mode
        """

    @View(calc_data, TIGraphedEquation[9])
    def X5T(self) -> TIGraphedEquation:
        """
        X5T: The 5th X-component in parametric mode
        """

    @View(calc_data, TIGraphedEquation[10])
    def Y5T(self) -> TIGraphedEquation:
        """
        Y5T: The 5th Y-component in parametric mode
        """

    @View(calc_data, TIGraphedEquation[11])
    def X6T(self) -> TIGraphedEquation:
        """
        X6T: The 6th X-component in parametric mode
        """

    @View(calc_data, TIGraphedEquation[12])
    def Y6T(self) -> TIGraphedEquation:
        """
        Y6T: The 6th Y-component in parametric mode
        """

    @Loader[dict]
    def load_dict(self, dct: dict = None):
        if dct is None:
            with open(os.path.join(os.path.dirname(__file__), "json/param.default.json")) as file:
                dct = json.load(file)

        super().load_dict(dct)

    def _load_dict(self, dct: dict):
        for i in range(1, self.num_styles + 1):
            if (x_style := getattr(self, f"X{i}T").style) != (y_style := getattr(self, f"Y{i}T").style):
                warn(f"X and Y component styles do not agree (X{i}T: {x_style}, Y{i}T: {y_style}).",
                     UserWarning)

            if (x_color := getattr(self, f"X{i}T").color) != (y_color := getattr(self, f"Y{i}T").color):
                warn(f"X and Y component colors do not agree (X{i}T: {x_color}, Y{i}T: {y_color}).",
                     UserWarning)

    def dict(self) -> dict:
        return super().dict() | {
            "specificData": {
                "settings": {
                    "Tmin": self.Tmin.json_number(),
                    "Tmax": self.Tmax.json_number(),
                    "Tstep": self.Tstep.json_number(),
                },
                "equations": {
                    equation.name: equation.dict() for equation in self.equations
                }
            }
        }


class TIParamGDB(TIGDB, TIMonoParamGDB):
    min_data_length = 144

    @Section()
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, String)[-14:-11]
    def color_magic(self) -> str:
        """
        Magic to identify the GDB as color-oriented -`84C`
        """

    def mono(self) -> TIMonoParamGDB:
        """
        :return: A copy of this GDB with all color data removed
        """

        return TIMonoParamGDB(self.bytes()[:-14])

    def _load_dict(self, dct: dict):
        super(TIGDB, self)._load_dict(dct)
        super(TIParamGDB, self)._load_dict(dct)

    def dict(self) -> dict:
        return super(TIGDB, self).dict() | super(TIParamGDB, self).dict()


class TIMonoPolarGDB(TIMonoGDB):
    mode_byte = 0x20

    min_data_length = 112

    num_equations = 6
    num_parameters = 3
    num_styles = 6

    equation_names = ["r1", "r2", "r3", "r4", "r5", "r6"]

    @Section()
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, Integer)[3:4]
    def mode_id(self) -> int:
        """
        The mode ID for the GDB - `0x20`
        """

    @View(calc_data, GraphRealEntry)[61:70]
    def Thetamin(self) -> GraphRealEntry:
        """
        θmin: The initial angle
        """

    @View(calc_data, GraphRealEntry)[70:79]
    def Thetamax(self) -> GraphRealEntry:
        """
        θmax: The final angle
        """

    @View(calc_data, GraphRealEntry)[79:88]
    def Thetastep(self) -> GraphRealEntry:
        """
        θstep: The angle increment
        """

    @View(calc_data, TIGraphedEquation[1])
    def r1(self) -> TIGraphedEquation:
        """
        r1: The 1st equation in polar mode
        """

    @View(calc_data, TIGraphedEquation[2])
    def r2(self) -> TIGraphedEquation:
        """
        r1: The 2nd equation in polar mode
        """

    @View(calc_data, TIGraphedEquation[3])
    def r3(self) -> TIGraphedEquation:
        """
        r3: The 3rd equation in polar mode
        """

    @View(calc_data, TIGraphedEquation[4])
    def r4(self) -> TIGraphedEquation:
        """
        r4: The 4th equation in polar mode
        """

    @View(calc_data, TIGraphedEquation[5])
    def r5(self) -> TIGraphedEquation:
        """
        r5: The 5th equation in polar mode
        """

    @View(calc_data, TIGraphedEquation[6])
    def r6(self) -> TIGraphedEquation:
        """
        r6: The 6th equation in polar mode
        """

    @Loader[dict]
    def load_dict(self, dct: dict = None):
        if dct is None:
            with open(os.path.join(os.path.dirname(__file__), "json/polar.default.json")) as file:
                dct = json.load(file)

        super().load_dict(dct)

    def dict(self) -> dict:
        return super().dict() | {
            "specificData": {
                "settings": {
                    "Thetamin": self.Thetamin.json_number(),
                    "Thetamax": self.Thetamax.json_number(),
                    "Thetastep": self.Thetastep.json_number(),
                },
                "equations": {
                    equation.name: equation.dict() for equation in self.equations
                }
            }
        }


class TIPolarGDB(TIGDB, TIMonoPolarGDB):
    min_data_length = 126

    @Section()
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, String)[-14:-11]
    def color_magic(self) -> str:
        """
        Magic to identify the GDB as color-oriented - `84C`
        """

    def mono(self) -> TIMonoPolarGDB:
        """
        :return: A copy of this GDB with all color data removed
        """

        return TIMonoPolarGDB(self.bytes()[:-14])

    def _load_dict(self, dct: dict):
        super(TIGDB, self)._load_dict(dct)
        super(TIPolarGDB, self)._load_dict(dct)

    def dict(self) -> dict:
        return super(TIGDB, self).dict() | super(TIPolarGDB, self).dict()


class TIMonoSeqGDB(TIMonoGDB):
    mode_byte = 0x80

    min_data_length = 163

    num_equations = 3
    num_parameters = 10
    num_styles = 3

    equation_names = ["u", "v", "w"]

    @Section()
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, Integer)[3:4]
    def mode_id(self) -> int:
        """
        The mode ID for the GDB - `0x80`
        """

    @View(calc_data, GraphMode)[5:6]
    def sequence_flags(self) -> SeqMode:
        """
        The flags for the sequence mode settings
        """

    @View(calc_data, GraphRealEntry)[61:70]
    def PlotStart(self, value) -> GraphRealEntry:
        """
        PlotStart: The initial value of 𝑛 for sequential plots

        The value must be an integer.
        """

        if int(value) != float(value):
            warn(f"Expected an integer, got {float(value)}.",
                 UserWarning)

        return value

    @View(calc_data, GraphRealEntry)[70:79]
    def nMax(self, value) -> GraphRealEntry:
        """
        𝑛Max: The final value of 𝑛

        The value must be an integer.
        """

        if int(value) != float(value):
            warn(f"Expected an integer, got {float(value)}.",
                 UserWarning)

        return value

    @View(calc_data, GraphRealEntry)[79:88]
    def unMin(self) -> GraphRealEntry:
        """
        u(𝑛Min): The initial value of u at 𝑛Min
        """

    @View(calc_data, GraphRealEntry)[88:97]
    def vnMin(self) -> GraphRealEntry:
        """
        v(𝑛Min): The initial value of v at 𝑛Min
        """

    @View(calc_data, GraphRealEntry)[97:106]
    def nMin(self, value) -> GraphRealEntry:
        """
        nMin: the initial value of 𝑛

        The value must be an integer.
        """

        if int(value) != float(value):
            warn(f"Expected an integer, got {float(value)}.",
                 UserWarning)

        return value

    @View(calc_data, GraphRealEntry)[106:115]
    def unMinp1(self) -> GraphRealEntry:
        """
        u(𝑛Min+1): The initial value of u at 𝑛Min + 1
        """

    @View(calc_data, GraphRealEntry)[115:124]
    def vnMinp1(self) -> GraphRealEntry:
        """
        v(nMin+1): The initial value of v at 𝑛Min + 1
        """

    @View(calc_data, GraphRealEntry)[124:133]
    def wnMin(self) -> GraphRealEntry:
        """
        w(𝑛Min): The initial value of w at 𝑛Min
        """

    @View(calc_data, GraphRealEntry)[133:142]
    def PlotStep(self, value) -> GraphRealEntry:
        """
        PlotStep: The 𝑛 increment for sequential plots

        The value must be an integer.
        """

        if int(value) != float(value):
            warn(f"Expected an integer, got {float(value)}.",
                 UserWarning)

        return value

    @View(calc_data, GraphRealEntry)[142:151]
    def wnMinp1(self) -> GraphRealEntry:
        """
        w(𝑛Min+1): The initial value of w at 𝑛Min + 1
        """

    @View(calc_data, TIGraphedEquation[1])
    def u(self) -> TIGraphedEquation:
        """
        u: The 1st equation in sequence mode
        """

    @View(calc_data, TIGraphedEquation[2])
    def v(self) -> TIGraphedEquation:
        """
        v: The 2nd equation in sequence mode
        """

    @View(calc_data, TIGraphedEquation[3])
    def w(self) -> TIGraphedEquation:
        """
        w: The 3rd equation in sequence mode
        """

    @Loader[dict]
    def load_dict(self, dct: dict = None):
        if dct is None:
            with open(os.path.join(os.path.dirname(__file__), "json/seq.default.json")) as file:
                dct = json.load(file)

        super().load_dict(dct)

    def _load_dict(self, dct: dict):
        ext_settings = dct.get("extSettings", {})
        if "seqMode" in ext_settings:
            match ext_settings["seqMode"]:
                case "SEQ(n)":
                    self.extended_mode_flags |= GraphMode.SEQ_n
                case "SEQ(n+1)":
                    self.extended_mode_flags |= GraphMode.SEQ_np1
                case "SEQ(n+2)":
                    self.extended_mode_flags |= GraphMode.SEQ_np2

        if "seqSettings" in dct:
            try:
                self.sequence_flags |= getattr(SeqMode, dct["seqMode"])
            except AttributeError:
                warn(f"Unrecognized sequence mode ({dct['seqMode']}).",
                     UserWarning)

    def dict(self) -> dict:
        dct = super().dict()

        match self.extended_mode_flags:
            case _SEQ_n if GraphMode.SEQ_n in self.extended_mode_flags:
                dct["extSettings"]["seqMode"] = "SEQ(n)"
            case _SEQ_np1 if GraphMode.SEQ_np1 in self.extended_mode_flags:
                dct["extSettings"]["seqMode"] = "SEQ(n+1)"
            case _SEQ_np2 if GraphMode.SEQ_np2 in self.extended_mode_flags:
                dct["extSettings"]["seqMode"] = "SEQ(n+2)"

        match self.sequence_flags:
            case _Time if SeqMode.Time in self.sequence_flags:
                dct["seqSettings"] = {"mode": "Time"}
            case _Web if SeqMode.Web in self.sequence_flags:
                dct["seqSettings"] = {"mode": "Web"}
            case _VertWeb if SeqMode.VertWeb in self.sequence_flags:
                dct["seqSettings"] = {"mode": "VertWeb"}
            case _uv if SeqMode.uv in self.sequence_flags:
                dct["seqSettings"] = {"mode": "uv"}
            case _vw if SeqMode.vw in self.sequence_flags:
                dct["seqSettings"] = {"mode": "vw"}
            case _uw if SeqMode.uw in self.sequence_flags:
                dct["seqSettings"] = {"mode": "uw"}

        return dct | {
            "specificData": {
                "settings": {
                    "nMin": int(self.nMin),
                    "nMax": int(self.nMax),
                    "PlotStart": int(self.PlotStart),
                    "PlotStep": int(self.PlotStep),
                    "unMin": self.unMin.json_number(),
                    "unMinp1": self.unMinp1.json_number(),
                    "vnMin": self.vnMinp1.json_number(),
                    "vnMinp1": self.vnMinp1.json_number(),
                    "wnMin": self.wnMin.json_number(),
                    "wnMinp1": self.wnMinp1.json_number()
                },
                "equations": {
                    equation.name: equation.dict() for equation in self.equations
                }
            }
        }


class TISeqGDB(TIGDB, TIMonoSeqGDB):
    min_data_length = 174

    @Section()
    def calc_data(self) -> bytearray:
        pass

    @View(calc_data, String)[-11:-8]
    def color_magic(self) -> str:
        """
        Magic to identify the GDB as color-oriented - `84C`
        """

    def mono(self) -> TIMonoSeqGDB:
        """
        :return: A copy of this GDB with all color data removed
        """

        return TIMonoSeqGDB(self.bytes()[:-11])

    def _load_dict(self, dct: dict):
        super(TIGDB, self)._load_dict(dct)
        super(TISeqGDB, self)._load_dict(dct)

    def dict(self) -> dict:
        return super(TIGDB, self).dict() | super(TISeqGDB, self).dict()


__all__ = ["TIMonoGDB", "TIMonoFuncGDB", "TIMonoParamGDB", "TIMonoPolarGDB", "TIMonoSeqGDB",
           "TIGDB", "TIFuncGDB", "TIParamGDB", "TIPolarGDB", "TISeqGDB",
           "TIGraphedEquation", "EquationFlags", "GraphMode", "GraphStyle", "GraphColor", "GlobalLineStyle"]



from typing import ByteString

from tivars.data import String
from ..models import *
from .tokens.scripts import *


def decode(bytestream: ByteString, *,
           tokens: Tokens = None, lang: str = "en",
           mode: str = "display") -> tuple[str | bytes, OsVersion]:
    tokens = tokens or TI_84PCE.tokens

    out = []
    since = OsVersions.INITIAL

    index = 0
    curr_bytes = b''
    while index < len(bytestream):
        curr_bytes += bytestream[index:][:1]

        if curr_bytes[0]:
            if curr_bytes in tokens.bytes:
                out.append(getattr(tokens.bytes[curr_bytes].langs[lang], mode))
                since = max(tokens.bytes[curr_bytes].since, since)

                curr_bytes = b''

            elif len(curr_bytes) >= 2:
                if not any(key.startswith(curr_bytes[:1]) for key in tokens.bytes):
                    raise ValueError(f"unrecognized byte '0x{curr_bytes[0]:x}' at position {index}")

                else:
                    raise ValueError(f"unrecognized bytes '0x{curr_bytes[0]:x}{curr_bytes[1]:x}' at position {index}")

        index += 1

    return b''.join(out) if mode == "ti_ascii" else "".join(out), since


def encode(string: str, trie: TokenTrie) -> tuple[bytes, OsVersion]:
    data = b''
    since = OsVersions.INITIAL
    index = 0

    while string:
        token, remainder = trie.get_longest_token(string)

        if token is None:
            raise ValueError(f"could not tokenize input at position {index}: '{string[:12]}'")

        data += token.bits
        since = max(token.since, since)

        index += len(string) - len(remainder)
        string = remainder

    return data, since


class TokenizedString(String):
    _T = str

    @classmethod
    def get(cls, data: bytes, **kwargs) -> _T:
        return decode(data.ljust(8, b'\x00'))[0]

    @classmethod
    def set(cls, value: _T, **kwargs) -> bytes:
        return encode(value, TI_84PCE.get_trie())[0].rstrip(b'\x00')


__all__ = ["decode", "encode", "TokenizedString",
           "Tokens", "OsVersion", "OsVersions"]
